#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.bindings.pdb import main


if __name__ == '__main__':
	main()



# vim:fileencoding=utf-8:noet
from __future__ import (division, absolute_import, print_function)

import argparse

import powerline.bindings.config as config


class StrFunction(object):
	def __init__(self, function, name=None):
		self.name = name or function.__name__
		self.function = function

	def __call__(self, *args, **kwargs):
		self.function(*args, **kwargs)

	def __str__(self):
		return self.name


TMUX_ACTIONS = {
	'source': StrFunction(config.source_tmux_files, 'source'),
	'setenv': StrFunction(config.init_tmux_environment, 'setenv'),
	'setup': StrFunction(config.tmux_setup, 'setup'),
}


SHELL_ACTIONS = {
	'command': StrFunction(config.shell_command, 'command'),
	'uses': StrFunction(config.uses),
}


class ConfigArgParser(argparse.ArgumentParser):
	def parse_args(self, *args, **kwargs):
		ret = super(ConfigArgParser, self).parse_args(*args, **kwargs)
		if not hasattr(ret, 'function'):
			# In Python-3* `powerline-config` (without arguments) raises 
			# AttributeError. I have not found any standard way to display same 
			# error message as in Python-2*.
			self.error('too few arguments')
		return ret


def get_argparser(ArgumentParser=ConfigArgParser):
	parser = ArgumentParser(description='Script used to obtain powerline configuration.')
	parser.add_argument(
		'-p', '--config-path', action='append', metavar='PATH',
		help='Path to configuration directory. If it is present '
		     'then configuration files will only be seeked in the provided path. '
		     'May be provided multiple times to search in a list of directories.'
	)
	subparsers = parser.add_subparsers()
	tmux_parser = subparsers.add_parser('tmux', help='Tmux-specific commands')
	tmux_parser.add_argument(
		'function',
		choices=tuple(TMUX_ACTIONS.values()),
		metavar='ACTION',
		type=(lambda v: TMUX_ACTIONS.get(v)),
		help='If action is `source\' then version-specific tmux configuration '
		     'files are sourced, if it is `setenv\' then special '
		     '(prefixed with `_POWERLINE\') tmux global environment variables '
		     'are filled with data from powerline configuration.'
	)

	shell_parser = subparsers.add_parser('shell', help='Shell-specific commands')
	shell_parser.add_argument(
		'function',
		choices=tuple(SHELL_ACTIONS.values()),
		type=(lambda v: SHELL_ACTIONS.get(v)),
		metavar='ACTION',
		help='If action is `command\' then preferred powerline command is '
		     'output, if it is `uses\' then powerline-config script will exit '
		     'with 1 if specified component is disabled and 0 otherwise.',
	)
	shell_parser.add_argument(
		'component',
		nargs='?',
		choices=('tmux', 'prompt'),
		metavar='COMPONENT',
		help='Only applicable for `uses\' subcommand: makes `powerline-config\' '
		     'exit with 0 if specific component is enabled and with 1 otherwise. '
		     '`tmux\' component stands for tmux bindings '
		     '(e.g. those that notify tmux about current directory changes), '
		     '`prompt\' component stands for shell prompt.'
	)
	shell_parser.add_argument(
		'-s', '--shell',
		metavar='SHELL',
		help='Shell for which query is run',
	)
	return parser



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import platform

from powerline.renderers.shell.readline import ReadlineRenderer
from powerline.renderer import Renderer


class PDBRenderer(ReadlineRenderer):
	'''PDB-specific powerline renderer
	'''
	pdb = None
	initial_stack_length = None

	def get_segment_info(self, segment_info, mode):
		r = self.segment_info.copy()
		r['pdb'] = self.pdb
		r['initial_stack_length'] = self.initial_stack_length
		r['curframe'] = self.pdb.curframe
		return r

	def set_pdb(self, pdb):
		'''Record currently used :py:class:`pdb.Pdb` instance

		Must be called before first calling :py:meth:`render` method.

		:param pdb.Pdb pdb:
			Used :py:class:`pdb.Pdb` instance. This instance will later be used 
			by :py:meth:`get_segment_info` for patching :ref:`segment_info 
			<dev-segments-info>` dictionary.
		'''
		self.pdb = pdb

	def render(self, **kwargs):
		if self.initial_stack_length is None:
			self.initial_stack_length = len(self.pdb.stack) - 1
		return Renderer.render(self, **kwargs)

	if sys.version_info < (3,) and platform.python_implementation() == 'PyPy':
		def do_render(self, **kwargs):
			# Make sure that only ASCII characters survive
			ret = super(PDBRenderer, self).do_render(**kwargs)
			ret = ret.encode('ascii', 'replace')
			ret = ret.decode('ascii')
			return ret


renderer = PDBRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell.zsh import ZshPromptRenderer


class TcshPromptRenderer(ZshPromptRenderer):
	'''Powerline tcsh prompt segment renderer.'''
	character_translations = ZshPromptRenderer.character_translations.copy()
	character_translations[ord('%')] = '%%'
	character_translations[ord('\\')] = '\\\\'
	character_translations[ord('^')] = '\\^'
	character_translations[ord('!')] = '\\!'


renderer = TcshPromptRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell import ShellRenderer


class BashPromptRenderer(ShellRenderer):
	'''Powerline bash prompt segment renderer.'''
	escape_hl_start = '\['
	escape_hl_end = '\]'

	character_translations = ShellRenderer.character_translations.copy()
	character_translations[ord('$')] = '\\$'
	character_translations[ord('`')] = '\\`'
	character_translations[ord('\\')] = '\\\\'


renderer = BashPromptRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import sys
import logging

from threading import Lock, Event

from powerline.colorscheme import Colorscheme
from powerline.lib.config import ConfigLoader
from powerline.lib.unicode import safe_unicode, FailedUnicode
from powerline.config import DEFAULT_SYSTEM_CONFIG_DIR
from powerline.lib.dict import mergedicts
from powerline.lib.encoding import get_preferred_output_encoding
from powerline.lib.path import join


class NotInterceptedError(BaseException):
	pass


def _config_loader_condition(path):
	if path and os.path.isfile(path):
		return path
	return None


def _find_config_files(search_paths, config_file, config_loader=None, loader_callback=None):
	config_file += '.json'
	found = False
	for path in search_paths:
		config_file_path = join(path, config_file)
		if os.path.isfile(config_file_path):
			yield config_file_path
			found = True
		elif config_loader:
			config_loader.register_missing(_config_loader_condition, loader_callback, config_file_path)
	if not found:
		raise IOError('Config file not found in search paths ({0}): {1}'.format(
			', '.join(search_paths),
			config_file
		))


class PowerlineLogger(object):
	'''Proxy class for logging.Logger instance

	It emits messages in format ``{ext}:{prefix}:{message}`` where

	``{ext}``
		is a used powerline extension (e.g. “vim”, “shell”, “ipython”).
	``{prefix}``
		is a local prefix, usually a segment name.
	``{message}``
		is the original message passed to one of the logging methods.

	Each of the methods (``critical``, ``exception``, ``info``, ``error``, 
	``warn``, ``debug``) expects to receive message in an ``str.format`` format, 
	not in printf-like format.

	Log is saved to the location :ref:`specified by user <config-common-log>`.
	'''

	def __init__(self, use_daemon_threads, logger, ext):
		self.logger = logger
		self.ext = ext
		self.use_daemon_threads = use_daemon_threads
		self.prefix = ''
		self.last_msgs = {}

	def _log(self, attr, msg, *args, **kwargs):
		prefix = kwargs.get('prefix') or self.prefix
		prefix = self.ext + ((':' + prefix) if prefix else '')
		msg = safe_unicode(msg)
		if args or kwargs:
			args = [safe_unicode(s) if isinstance(s, bytes) else s for s in args]
			kwargs = dict((
				(k, safe_unicode(v) if isinstance(v, bytes) else v)
				for k, v in kwargs.items()
			))
			msg = msg.format(*args, **kwargs)
		msg = prefix + ':' + msg
		key = attr + ':' + prefix
		if msg != self.last_msgs.get(key):
			getattr(self.logger, attr)(msg)
			self.last_msgs[key] = msg

	def critical(self, msg, *args, **kwargs):
		self._log('critical', msg, *args, **kwargs)

	def exception(self, msg, *args, **kwargs):
		self._log('exception', msg, *args, **kwargs)

	def info(self, msg, *args, **kwargs):
		self._log('info', msg, *args, **kwargs)

	def error(self, msg, *args, **kwargs):
		self._log('error', msg, *args, **kwargs)

	def warn(self, msg, *args, **kwargs):
		self._log('warning', msg, *args, **kwargs)

	def debug(self, msg, *args, **kwargs):
		self._log('debug', msg, *args, **kwargs)


_fallback_logger = None


def get_fallback_logger(stream=None):
	global _fallback_logger
	if _fallback_logger:
		return _fallback_logger

	log_format = '%(asctime)s:%(levelname)s:%(message)s'
	formatter = logging.Formatter(log_format)

	level = logging.WARNING
	handler = logging.StreamHandler(stream)
	handler.setLevel(level)
	handler.setFormatter(formatter)

	logger = logging.getLogger('powerline')
	logger.setLevel(level)
	logger.addHandler(handler)
	_fallback_logger = PowerlineLogger(None, logger, '_fallback_')
	return _fallback_logger


def _generate_change_callback(lock, key, dictionary):
	def on_file_change(path):
		with lock:
			dictionary[key] = True
	return on_file_change


def get_config_paths():
	'''Get configuration paths from environment variables.

	Uses $XDG_CONFIG_HOME and $XDG_CONFIG_DIRS according to the XDG specification.

	:return: list of paths
	'''
	config_home = os.environ.get('XDG_CONFIG_HOME', os.path.join(os.path.expanduser('~'), '.config'))
	config_path = join(config_home, 'powerline')
	config_paths = [config_path]
	config_dirs = os.environ.get('XDG_CONFIG_DIRS', DEFAULT_SYSTEM_CONFIG_DIR)
	if config_dirs is not None:
		config_paths[:0] = reversed([join(d, 'powerline') for d in config_dirs.split(':')])
	plugin_path = join(os.path.realpath(os.path.dirname(__file__)), 'config_files')
	config_paths.insert(0, plugin_path)
	return config_paths


def generate_config_finder(get_config_paths=get_config_paths):
	'''Generate find_config_files function

	This function will find .json file given its path.

	:param function get_config_paths:
		Function that being called with no arguments will return a list of paths 
		that should be searched for configuration files.

	:return:
		Function that being given configuration file name will return full path 
		to it or raise IOError if it failed to find the file.
	'''
	config_paths = get_config_paths()
	return lambda *args: _find_config_files(config_paths, *args)


def load_config(cfg_path, find_config_files, config_loader, loader_callback=None):
	'''Load configuration file and setup watches

	Watches are only set up if loader_callback is not None.

	:param str cfg_path:
		Path for configuration file that should be loaded.
	:param function find_config_files:
		Function that finds configuration file. Check out the description of 
		the return value of ``generate_config_finder`` function.
	:param ConfigLoader config_loader:
		Configuration file loader class instance.
	:param function loader_callback:
		Function that will be called by config_loader when change to 
		configuration file is detected.

	:return: Configuration file contents.
	'''
	found_files = find_config_files(cfg_path, config_loader, loader_callback)
	ret = None
	for path in found_files:
		if loader_callback:
			config_loader.register(loader_callback, path)
		if ret is None:
			ret = config_loader.load(path)
		else:
			mergedicts(ret, config_loader.load(path))
	return ret


def _get_log_handler(common_config, stream=None):
	'''Get log handler.

	:param dict common_config:
		Configuration dictionary used to create handler.

	:return: logging.Handler subclass.
	'''
	log_file = common_config['log_file']
	if log_file:
		log_file = os.path.expanduser(log_file)
		log_dir = os.path.dirname(log_file)
		if not os.path.isdir(log_dir):
			os.mkdir(log_dir)
		return logging.FileHandler(log_file)
	else:
		return logging.StreamHandler(stream)


def create_logger(common_config, stream=None):
	'''Create logger according to provided configuration
	'''
	log_format = common_config['log_format']
	formatter = logging.Formatter(log_format)

	level = getattr(logging, common_config['log_level'])
	handler = _get_log_handler(common_config, stream)
	handler.setLevel(level)
	handler.setFormatter(formatter)

	logger = logging.getLogger('powerline')
	logger.setLevel(level)
	logger.addHandler(handler)
	return logger


def finish_common_config(encoding, common_config):
	'''Add default values to common config and expand ~ in paths

	:param dict common_config:
		Common configuration, as it was just loaded.

	:return:
		Copy of common configuration with all configuration keys and expanded 
		paths.
	'''
	encoding = encoding.lower()
	if encoding.startswith('utf') or encoding.startswith('ucs'):
		default_top_theme = 'powerline'
	else:
		default_top_theme = 'ascii'

	common_config = common_config.copy()
	common_config.setdefault('default_top_theme', default_top_theme)
	common_config.setdefault('paths', [])
	common_config.setdefault('watcher', 'auto')
	common_config.setdefault('log_level', 'WARNING')
	common_config.setdefault('log_format', '%(asctime)s:%(levelname)s:%(message)s')
	common_config.setdefault('term_truecolor', False)
	common_config.setdefault('term_escape_style', 'auto')
	common_config.setdefault('ambiwidth', 1)
	common_config.setdefault('additional_escapes', None)
	common_config.setdefault('reload_config', True)
	common_config.setdefault('interval', None)
	common_config.setdefault('log_file', None)

	common_config['paths'] = [
		os.path.expanduser(path) for path in common_config['paths']
	]

	return common_config


if sys.version_info < (3,):
	# `raise exception[0], None, exception[1]` is a SyntaxError in python-3*
	# Not using ('''…''') because this syntax does not work in python-2.6
	exec((
		'def reraise(exception):\n'
		'	if type(exception) is tuple:\n'
		'		raise exception[0], None, exception[1]\n'
		'	else:\n'
		'		raise exception\n'
	))
else:
	def reraise(exception):
		if type(exception) is tuple:
			raise exception[0].with_traceback(exception[1])
		else:
			raise exception


def gen_module_attr_getter(pl, import_paths, imported_modules):
	def get_module_attr(module, attr, prefix='powerline'):
		'''Import module and get its attribute.

		Replaces ``from {module} import {attr}``.

		:param str module:
			Module name, will be passed as first argument to ``__import__``.
		:param str attr:
			Module attribute, will be passed to ``__import__`` as the only value 
			in ``fromlist`` tuple.

		:return:
			Attribute value or ``None``. Note: there is no way to distinguish 
			between successfull import of attribute equal to ``None`` and 
			unsuccessfull import.
		'''
		oldpath = sys.path
		sys.path = import_paths + sys.path
		module = str(module)
		attr = str(attr)
		try:
			imported_modules.add(module)
			return getattr(__import__(module, fromlist=(attr,)), attr)
		except Exception as e:
			pl.exception('Failed to import attr {0} from module {1}: {2}', attr, module, str(e), prefix=prefix)
			return None
		finally:
			sys.path = oldpath

	return get_module_attr


class Powerline(object):
	'''Main powerline class, entrance point for all powerline uses. Sets 
	powerline up and loads the configuration.

	:param str ext:
		extension used. Determines where configuration files will 
		searched and what renderer module will be used. Affected: used ``ext`` 
		dictionary from :file:`powerline/config.json`, location of themes and 
		colorschemes, render module (``powerline.renders.{ext}``).
	:param str renderer_module:
		Overrides renderer module (defaults to ``ext``). Should be the name of 
		the package imported like this: ``powerline.renderers.{render_module}``. 
		If this parameter contains a dot ``powerline.renderers.`` is not 
		prepended. There is also a special case for renderers defined in 
		toplevel modules: ``foo.`` (note: dot at the end) tries to get renderer 
		from module ``foo`` (because ``foo`` (without dot) tries to get renderer 
		from module ``powerline.renderers.foo``). When ``.foo`` (with leading 
		dot) variant is used ``renderer_module`` will be 
		``powerline.renderers.{ext}{renderer_module}``.
	:param bool run_once:
		Determines whether :py:meth:`render` method will be run only once 
		during python session.
	:param Logger logger:
		If present no new logger will be created and the provided logger will be 
		used.
	:param bool use_daemon_threads:
		When creating threads make them daemon ones.
	:param Event shutdown_event:
		Use this Event as shutdown_event instead of creating new event.
	:param ConfigLoader config_loader:
		Instance of the class that manages (re)loading of the configuration.
	'''

	def __init__(self, *args, **kwargs):
		self.init_args = (args, kwargs)
		self.init(*args, **kwargs)

	def init(self,
	         ext,
	         renderer_module=None,
	         run_once=False,
	         logger=None,
	         use_daemon_threads=True,
	         shutdown_event=None,
	         config_loader=None):
		'''Do actual initialization.

		__init__ function only stores the arguments and runs this function. This 
		function exists for powerline to be able to reload itself: it is easier 
		to make ``__init__`` store arguments and call overriddable ``init`` than 
		tell developers that each time they override Powerline.__init__ in 
		subclasses they must store actual arguments.
		'''
		self.ext = ext
		self.run_once = run_once
		self.logger = logger
		self.use_daemon_threads = use_daemon_threads

		if not renderer_module:
			self.renderer_module = 'powerline.renderers.' + ext
		elif '.' not in renderer_module:
			self.renderer_module = 'powerline.renderers.' + renderer_module
		elif renderer_module.startswith('.'):
			self.renderer_module = 'powerline.renderers.' + ext + renderer_module
		elif renderer_module.endswith('.'):
			self.renderer_module = renderer_module[:-1]
		else:
			self.renderer_module = renderer_module

		self.find_config_files = generate_config_finder(self.get_config_paths)

		self.cr_kwargs_lock = Lock()
		self.cr_kwargs = {}
		self.cr_callbacks = {}
		for key in ('main', 'colors', 'colorscheme', 'theme'):
			self.cr_kwargs['load_' + key] = True
			self.cr_callbacks[key] = _generate_change_callback(
				self.cr_kwargs_lock,
				'load_' + key,
				self.cr_kwargs
			)

		self.shutdown_event = shutdown_event or Event()
		self.config_loader = config_loader or ConfigLoader(shutdown_event=self.shutdown_event, run_once=run_once)
		self.run_loader_update = False

		self.renderer_options = {}

		self.prev_common_config = None
		self.prev_ext_config = None
		self.pl = None
		self.setup_args = ()
		self.setup_kwargs = {}
		self.imported_modules = set()

	get_encoding = staticmethod(get_preferred_output_encoding)
	'''Get encoding used by the current application

	Usually returns encoding of the current locale.
	'''

	def create_logger(self):
		'''Create logger

		This function is used to create logger unless it was already specified 
		at initialization.
		'''
		return create_logger(self.common_config, self.default_log_stream)

	def create_renderer(self, load_main=False, load_colors=False, load_colorscheme=False, load_theme=False):
		'''(Re)create renderer object. Can be used after Powerline object was 
		successfully initialized. If any of the below parameters except 
		``load_main`` is True renderer object will be recreated.

		:param bool load_main:
			Determines whether main configuration file (:file:`config.json`) 
			should be loaded. If appropriate configuration changes implies 
			``load_colorscheme`` and ``load_theme`` and recreation of renderer 
			object. Won’t trigger recreation if only unrelated configuration 
			changed.
		:param bool load_colors:
			Determines whether colors configuration from :file:`colors.json` 
			should be (re)loaded.
		:param bool load_colorscheme:
			Determines whether colorscheme configuration should be (re)loaded.
		:param bool load_theme:
			Determines whether theme configuration should be reloaded.
		'''
		common_config_differs = False
		ext_config_differs = False
		if load_main:
			self._purge_configs('main')
			config = self.load_main_config()
			self.common_config = finish_common_config(self.get_encoding(), config['common'])
			if self.common_config != self.prev_common_config:
				common_config_differs = True

				load_theme = (load_theme
					or not self.prev_common_config
					or self.prev_common_config['default_top_theme'] != self.common_config['default_top_theme'])

				self.prev_common_config = self.common_config

				self.import_paths = self.common_config['paths']

				if not self.logger:
					self.logger = self.create_logger()

				if not self.pl:
					self.pl = PowerlineLogger(self.use_daemon_threads, self.logger, self.ext)
					self.config_loader.pl = self.pl

				if not self.run_once:
					self.config_loader.set_watcher(self.common_config['watcher'])

				self.get_module_attr = gen_module_attr_getter(self.pl, self.import_paths, self.imported_modules)

				mergedicts(self.renderer_options, dict(
					pl=self.pl,
					term_truecolor=self.common_config['term_truecolor'],
					term_escape_style=self.common_config['term_escape_style'],
					ambiwidth=self.common_config['ambiwidth'],
					tmux_escape=self.common_config['additional_escapes'] == 'tmux',
					screen_escape=self.common_config['additional_escapes'] == 'screen',
					theme_kwargs={
						'ext': self.ext,
						'common_config': self.common_config,
						'run_once': self.run_once,
						'shutdown_event': self.shutdown_event,
						'get_module_attr': self.get_module_attr,
					},
				))

				if not self.run_once and self.common_config['reload_config']:
					interval = self.common_config['interval']
					self.config_loader.set_interval(interval)
					self.run_loader_update = (interval is None)
					if interval is not None and not self.config_loader.is_alive():
						self.config_loader.start()

			self.ext_config = config['ext'][self.ext]

			top_theme = (
				self.ext_config.get('top_theme')
				or self.common_config['default_top_theme']
			)
			self.theme_levels = (
				os.path.join('themes', top_theme),
				os.path.join('themes', self.ext, '__main__'),
			)
			self.renderer_options['theme_kwargs']['top_theme'] = top_theme

			if self.ext_config != self.prev_ext_config:
				ext_config_differs = True
				if (
					not self.prev_ext_config
					or self.ext_config.get('components') != self.prev_ext_config.get('components')
				):
					self.setup_components(self.ext_config.get('components'))
				if (
					not self.prev_ext_config
					or self.ext_config.get('local_themes') != self.prev_ext_config.get('local_themes')
				):
					self.renderer_options['local_themes'] = self.get_local_themes(self.ext_config.get('local_themes'))
				load_colorscheme = (
					load_colorscheme
					or not self.prev_ext_config
					or self.prev_ext_config['colorscheme'] != self.ext_config['colorscheme']
				)
				load_theme = (
					load_theme
					or not self.prev_ext_config
					or self.prev_ext_config['theme'] != self.ext_config['theme']
				)
				self.prev_ext_config = self.ext_config

		create_renderer = load_colors or load_colorscheme or load_theme or common_config_differs or ext_config_differs

		if load_colors:
			self._purge_configs('colors')
			self.colors_config = self.load_colors_config()

		if load_colorscheme or load_colors:
			self._purge_configs('colorscheme')
			if load_colorscheme:
				self.colorscheme_config = self.load_colorscheme_config(self.ext_config['colorscheme'])
			self.renderer_options['theme_kwargs']['colorscheme'] = (
				Colorscheme(self.colorscheme_config, self.colors_config))

		if load_theme:
			self._purge_configs('theme')
			self.renderer_options['theme_config'] = self.load_theme_config(self.ext_config.get('theme', 'default'))

		if create_renderer:
			Renderer = self.get_module_attr(self.renderer_module, 'renderer')
			if not Renderer:
				if hasattr(self, 'renderer'):
					return
				else:
					raise ImportError('Failed to obtain renderer')

			# Renderer updates configuration file via segments’ .startup thus it 
			# should be locked to prevent state when configuration was updated, 
			# but .render still uses old renderer.
			try:
				renderer = Renderer(**self.renderer_options)
			except Exception as e:
				self.exception('Failed to construct renderer object: {0}', str(e))
				if not hasattr(self, 'renderer'):
					raise
			else:
				self.renderer = renderer

	default_log_stream = sys.stdout
	'''Default stream for default log handler

	Usually it is ``sys.stderr``, but there is sometimes a reason to prefer 
	``sys.stdout`` or a custom file-like object. It is not supposed to be used 
	to write to some file.
	'''

	def setup_components(self, components):
		'''Run component-specific setup

		:param set components:
			Set of the enabled componets or None.

		Should be overridden by subclasses.
		'''
		pass

	@staticmethod
	def get_config_paths():
		'''Get configuration paths.

		Should be overridden in subclasses in order to provide a way to override 
		used paths.

		:return: list of paths
		'''
		return get_config_paths()

	def load_config(self, cfg_path, cfg_type):
		'''Load configuration and setup watches

		:param str cfg_path:
			Path to the configuration file without any powerline configuration 
			directory or ``.json`` suffix.
		:param str cfg_type:
			Configuration type. May be one of ``main`` (for ``config.json`` 
			file), ``colors``, ``colorscheme``, ``theme``.

		:return: dictionary with loaded configuration.
		'''
		return load_config(
			cfg_path,
			self.find_config_files,
			self.config_loader,
			self.cr_callbacks[cfg_type]
		)

	def _purge_configs(self, cfg_type):
		function = self.cr_callbacks[cfg_type]
		self.config_loader.unregister_functions(set((function,)))
		self.config_loader.unregister_missing(set(((self.find_config_files, function),)))

	def load_main_config(self):
		'''Get top-level configuration.

		:return: dictionary with :ref:`top-level configuration <config-main>`.
		'''
		return self.load_config('config', 'main')

	def _load_hierarhical_config(self, cfg_type, levels, ignore_levels):
		'''Load and merge multiple configuration files

		:param str cfg_type:
			Type of the loaded configuration files (e.g. ``colorscheme``, 
			``theme``).
		:param list levels:
			Configuration names resembling levels in hierarchy, sorted by 
			priority. Configuration file names with higher priority should go 
			last.
		:param set ignore_levels:
			If only files listed in this variable are present then configuration 
			file is considered not loaded: at least one file on the level not 
			listed in this variable must be present.
		'''
		config = {}
		loaded = 0
		exceptions = []
		for i, cfg_path in enumerate(levels):
			try:
				lvl_config = self.load_config(cfg_path, cfg_type)
			except IOError as e:
				if sys.version_info < (3,):
					tb = sys.exc_info()[2]
					exceptions.append((e, tb))
				else:
					exceptions.append(e)
			else:
				if i not in ignore_levels:
					loaded += 1
				mergedicts(config, lvl_config)
		if not loaded:
			for exception in exceptions:
				if type(exception) is tuple:
					e = exception[0]
				else:
					e = exception
				self.exception('Failed to load %s: {0}' % cfg_type, e, exception=exception)
			raise e
		return config

	def load_colorscheme_config(self, name):
		'''Get colorscheme.

		:param str name:
			Name of the colorscheme to load.

		:return: dictionary with :ref:`colorscheme configuration <config-colorschemes>`.
		'''
		levels = (
			os.path.join('colorschemes', name),
			os.path.join('colorschemes', self.ext, '__main__'),
			os.path.join('colorschemes', self.ext, name),
		)
		return self._load_hierarhical_config('colorscheme', levels, (1,))

	def load_theme_config(self, name):
		'''Get theme configuration.

		:param str name:
			Name of the theme to load.

		:return: dictionary with :ref:`theme configuration <config-themes>`
		'''
		levels = self.theme_levels + (
			os.path.join('themes', self.ext, name),
		)
		return self._load_hierarhical_config('theme', levels, (0, 1,))

	def load_colors_config(self):
		'''Get colorscheme.

		:return: dictionary with :ref:`colors configuration <config-colors>`.
		'''
		return self.load_config('colors', 'colors')

	@staticmethod
	def get_local_themes(local_themes):
		'''Get local themes. No-op here, to be overridden in subclasses if 
		required.

		:param dict local_themes:
			Usually accepts ``{matcher_name : theme_name}``. May also receive 
			None in case there is no local_themes configuration.

		:return:
			anything accepted by ``self.renderer.get_theme`` and processable by 
			``self.renderer.add_local_theme``. Renderer module is determined by 
			``__init__`` arguments, refer to its documentation.
		'''
		return None

	def update_renderer(self):
		'''Updates/creates a renderer if needed.'''
		if self.run_loader_update:
			self.config_loader.update()
		cr_kwargs = None
		with self.cr_kwargs_lock:
			if self.cr_kwargs:
				cr_kwargs = self.cr_kwargs.copy()
		if cr_kwargs:
			try:
				self.create_renderer(**cr_kwargs)
			except Exception as e:
				self.exception('Failed to create renderer: {0}', str(e))
				if hasattr(self, 'renderer'):
					with self.cr_kwargs_lock:
						self.cr_kwargs.clear()
				else:
					raise
			else:
				with self.cr_kwargs_lock:
					self.cr_kwargs.clear()

	def render(self, *args, **kwargs):
		'''Update/create renderer if needed and pass all arguments further to 
		``self.renderer.render()``.
		'''
		try:
			self.update_renderer()
			return self.renderer.render(*args, **kwargs)
		except Exception as e:
			exc = e
			try:
				self.exception('Failed to render: {0}', str(e))
			except Exception as e:
				exc = e
			ret = FailedUnicode(safe_unicode(exc))
			if kwargs.get('output_width', False):
				ret = ret, len(ret)
			return ret

	def render_above_lines(self, *args, **kwargs):
		'''Like .render(), but for ``self.renderer.render_above_lines()``
		'''
		try:
			self.update_renderer()
			for line in self.renderer.render_above_lines(*args, **kwargs):
				yield line
		except Exception as e:
			exc = e
			try:
				self.exception('Failed to render: {0}', str(e))
			except Exception as e:
				exc = e
			yield FailedUnicode(safe_unicode(exc))

	def setup(self, *args, **kwargs):
		'''Setup the environment to use powerline.

		Must not be overridden by subclasses. This one only saves setup 
		arguments for :py:meth:`reload` method and calls :py:meth:`do_setup`.
		'''
		self.shutdown_event.clear()
		self.setup_args = args
		self.setup_kwargs.update(kwargs)
		self.do_setup(*args, **kwargs)

	@staticmethod
	def do_setup():
		'''Function that does initialization

		Should be overridden by subclasses. May accept any number of regular or 
		keyword arguments.
		'''
		pass

	def reload(self):
		'''Reload powerline after update.

		Should handle most (but not all) powerline updates.

		Purges out all powerline modules and modules imported by powerline for 
		segment and matcher functions. Requires defining ``setup`` function that 
		updates reference to main powerline object.

		.. warning::
			Not guaranteed to work properly, use it at your own risk. It 
			may break your python code.
		'''
		import sys
		modules = self.imported_modules | set((module for module in sys.modules if module.startswith('powerline')))
		modules_holder = []
		for module in modules:
			try:
				# Needs to hold module to prevent garbage collecting until they 
				# are all reloaded.
				modules_holder.append(sys.modules.pop(module))
			except KeyError:
				pass
		PowerlineClass = getattr(__import__(self.__module__, fromlist=(self.__class__.__name__,)), self.__class__.__name__)
		self.shutdown(set_event=True)
		init_args, init_kwargs = self.init_args
		powerline = PowerlineClass(*init_args, **init_kwargs)
		powerline.setup(*self.setup_args, **self.setup_kwargs)

	def shutdown(self, set_event=True):
		'''Shut down all background threads.

		:param bool set_event:
			Set ``shutdown_event`` and call ``renderer.shutdown`` which should 
			shut down all threads. Set it to False unless you are exiting an 
			application.

			If set to False this does nothing more then resolving reference 
			cycle ``powerline → config_loader → bound methods → powerline`` by 
			unsubscribing from config_loader events.
		'''
		if set_event:
			self.shutdown_event.set()
			try:
				self.renderer.shutdown()
			except AttributeError:
				pass
		functions = tuple(self.cr_callbacks.values())
		self.config_loader.unregister_functions(set(functions))
		self.config_loader.unregister_missing(set(((self.find_config_files, function) for function in functions)))

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.shutdown()

	def exception(self, msg, *args, **kwargs):
		if 'prefix' not in kwargs:
			kwargs['prefix'] = 'powerline'
		exception = kwargs.pop('exception', None)
		pl = getattr(self, 'pl', None) or get_fallback_logger(self.default_log_stream)
		if exception:
			try:
				reraise(exception)
			except Exception:
				return pl.exception(msg, *args, **kwargs)
		return pl.exception(msg, *args, **kwargs)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.bindings.vim import list_tabpages


def single_tab(pl, segment_info, mode):
	'''Returns True if Vim has only one tab opened
	'''
	return len(list_tabpages()) == 1



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import re
import os
import subprocess

from collections import namedtuple

from powerline.lib.shell import run_cmd


TmuxVersionInfo = namedtuple('TmuxVersionInfo', ('major', 'minor', 'suffix'))


def get_tmux_executable_name():
	'''Returns tmux executable name

	It should be defined in POWERLINE_TMUX_EXE environment variable, otherwise 
	it is simply “tmux”.
	'''

	return os.environ.get('POWERLINE_TMUX_EXE', 'tmux')


def _run_tmux(runner, args):
	return runner([get_tmux_executable_name()] + list(args))


def run_tmux_command(*args):
	'''Run tmux command, ignoring the output'''
	_run_tmux(subprocess.check_call, args)


def get_tmux_output(pl, *args):
	'''Run tmux command and return its output'''
	return _run_tmux(lambda cmd: run_cmd(pl, cmd), args)


def set_tmux_environment(varname, value, remove=True):
	'''Set tmux global environment variable

	:param str varname:
		Name of the variable to set.
	:param str value:
		Variable value.
	:param bool remove:
		True if variable should be removed from the environment prior to 
		attaching any client (runs ``tmux set-environment -r {varname}``).
	'''
	run_tmux_command('set-environment', '-g', varname, value)
	if remove:
		run_tmux_command('set-environment', '-r', varname)


NON_DIGITS = re.compile('[^0-9]+')
DIGITS = re.compile('[0-9]+')
NON_LETTERS = re.compile('[^a-z]+')


def get_tmux_version(pl):
	version_string = get_tmux_output(pl, '-V')
	_, version_string = version_string.split(' ')
	version_string = version_string.strip()
	major, minor = version_string.split('.')
	suffix = DIGITS.subn('', minor)[0] or None
	minor = NON_DIGITS.subn('', minor)[0]
	return TmuxVersionInfo(int(major), int(minor), suffix)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from inspect import formatargspec

from sphinx.ext import autodoc

from powerline.lint.inspect import getconfigargspec
from powerline.segments import Segment
from powerline.lib.unicode import unicode


def formatvalue(val):
	if type(val) is str:
		return '="' + unicode(val, 'utf-8').replace('"', '\\"').replace('\\', '\\\\') + '"'
	else:
		return '=' + repr(val)


class ThreadedDocumenter(autodoc.FunctionDocumenter):
	'''Specialized documenter subclass for ThreadedSegment subclasses.'''
	@classmethod
	def can_document_member(cls, member, membername, isattr, parent):
		return (isinstance(member, Segment) or
			super(ThreadedDocumenter, cls).can_document_member(member, membername, isattr, parent))

	def format_args(self):
		argspec = getconfigargspec(self.object)
		return formatargspec(*argspec, formatvalue=formatvalue).replace('\\', '\\\\')


def setup(app):
	autodoc.setup(app)
	app.add_autodocumenter(ThreadedDocumenter)



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys

from time import sleep
from subprocess import Popen, PIPE

from powerline import Powerline
from powerline.lib.monotonic import monotonic

powerline = Powerline('wm', renderer_module='pango_markup')
powerline.update_renderer()

try:
	interval = float(sys.argv[1])
except IndexError:
	interval = 2


def read_to_log(pl, client):
	for line in client.stdout:
		if line:
			pl.info(line, prefix='awesome-client')
	for line in client.stderr:
		if line:
			pl.error(line, prefix='awesome-client')
	if client.wait():
		pl.error('Client exited with {0}', client.returncode, prefix='awesome')


while True:
	start_time = monotonic()
	s = powerline.render(side='right')
	request = 'powerline_widget:set_markup(\'' + s.replace('\\', '\\\\').replace('\'', '\\\'') + '\')\n'
	client = Popen(['awesome-client'], shell=False, stdout=PIPE, stderr=PIPE, stdin=PIPE)
	client.stdin.write(request.encode('utf-8'))
	client.stdin.close()
	read_to_log(powerline.pl, client)
	sleep(max(interval - (monotonic() - start_time), 0.1))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from datetime import datetime


def date(pl, format='%Y-%m-%d', istime=False):
	'''Return the current date.

	:param str format:
		strftime-style date format string
	:param bool istime:
		If true then segment uses ``time`` highlight group.

	Divider highlight group used: ``time:divider``.

	Highlight groups used: ``time`` or ``date``.
	'''
	return [{
		'contents': datetime.now().strftime(format),
		'highlight_groups': (['time'] if istime else []) + ['date'],
		'divider_highlight_group': 'time:divider' if istime else None,
	}]


UNICODE_TEXT_TRANSLATION = {
	ord('\''): '’',
	ord('-'): '‐',
}


def fuzzy_time(pl, unicode_text=False):
	'''Display the current time as fuzzy time, e.g. "quarter past six".

	:param bool unicode_text:
		If true then hyphenminuses (regular ASCII ``-``) and single quotes are 
		replaced with unicode dashes and apostrophes.
	'''
	hour_str = ['twelve', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten', 'eleven']
	minute_str = {
		5: 'five past',
		10: 'ten past',
		15: 'quarter past',
		20: 'twenty past',
		25: 'twenty-five past',
		30: 'half past',
		35: 'twenty-five to',
		40: 'twenty to',
		45: 'quarter to',
		50: 'ten to',
		55: 'five to',
	}
	special_case_str = {
		(23, 58): 'round about midnight',
		(23, 59): 'round about midnight',
		(0, 0): 'midnight',
		(0, 1): 'round about midnight',
		(0, 2): 'round about midnight',
		(12, 0): 'noon',
	}

	now = datetime.now()

	try:
		return special_case_str[(now.hour, now.minute)]
	except KeyError:
		pass

	hour = now.hour
	if now.minute > 32:
		if hour == 23:
			hour = 0
		else:
			hour += 1
	if hour > 11:
		hour = hour - 12
	hour = hour_str[hour]

	minute = int(round(now.minute / 5.0) * 5)
	if minute == 60 or minute == 0:
		result = ' '.join([hour, 'o\'clock'])
	else:
		minute = minute_str[minute]
		result = ' '.join([minute, hour])

	if unicode_text:
		result = result.translate(UNICODE_TEXT_TRANSLATION)

	return result



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import re

from powerline.bindings.vim import buffer_name


NERD_TREE_RE = re.compile(b'NERD_tree_\\d+')


def nerdtree(matcher_info):
	name = buffer_name(matcher_info)
	return name and NERD_TREE_RE.match(os.path.basename(name))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import codecs

try:
	import vim
except ImportError:
	vim = object()

from powerline.lib.unicode import unicode


if (
	hasattr(vim, 'options')
	and hasattr(vim, 'vvars')
	and vim.vvars['version'] > 703
):
	if sys.version_info < (3,):
		def get_vim_encoding():
			return vim.options['encoding'] or 'ascii'
	else:
		def get_vim_encoding():
			return vim.options['encoding'].decode('ascii') or 'ascii'
elif hasattr(vim, 'eval'):
	def get_vim_encoding():
		return vim.eval('&encoding') or 'ascii'
else:
	def get_vim_encoding():
		return 'utf-8'

get_vim_encoding.__doc__ = (
	'''Get encoding used for Vim strings

	:return:
		Value of ``&encoding``. If it is empty (i.e. Vim is compiled 
		without +multibyte) returns ``'ascii'``. When building documentation 
		outputs ``'utf-8'`` unconditionally.
	'''
)


vim_encoding = get_vim_encoding()


python_to_vim_types = {
	unicode: (
		lambda o: b'\'' + (o.translate({
			ord('\''): '\'\'',
		}).encode(vim_encoding)) + b'\''
	),
	list: (
		lambda o: b'[' + (
			b','.join((python_to_vim(i) for i in o))
		) + b']'
	),
	bytes: (lambda o: b'\'' + o.replace(b'\'', b'\'\'') + b'\''),
	int: (str if str is bytes else (lambda o: unicode(o).encode('ascii'))),
}
python_to_vim_types[float] = python_to_vim_types[int]


def python_to_vim(o):
	return python_to_vim_types[type(o)](o)


if sys.version_info < (3,):
	def str_to_bytes(s):
		return s

	def unicode_eval(expr):
		ret = vim.eval(expr)
		return ret.decode(vim_encoding, 'powerline_vim_strtrans_error')
else:
	def str_to_bytes(s):
		return s.encode(vim_encoding)

	def unicode_eval(expr):
		return vim.eval(expr)


def safe_bytes_eval(expr):
	return bytes(bytearray((
		int(chunk) for chunk in (
			vim.eval(
				b'substitute(' + expr + b', ' +
				b'\'^.*$\', \'\\=join(map(range(len(submatch(0))), ' +
				b'"char2nr(submatch(0)[v:val])"))\', "")'
			).split()
		)
	)))


def eval_bytes(expr):
	try:
		return str_to_bytes(vim.eval(expr))
	except UnicodeDecodeError:
		return safe_bytes_eval(expr)


def eval_unicode(expr):
	try:
		return unicode_eval(expr)
	except UnicodeDecodeError:
		return safe_bytes_eval(expr).decode(vim_encoding, 'powerline_vim_strtrans_error')


if hasattr(vim, 'bindeval'):
	rettype_func = {
		None: lambda f: f,
		'unicode': (
			lambda f: (
				lambda *args, **kwargs: (
					f(*args, **kwargs).decode(
						vim_encoding, 'powerline_vim_strtrans_error'
					))))
	}
	rettype_func['int'] = rettype_func['bytes'] = rettype_func[None]
	rettype_func['str'] = rettype_func['bytes'] if str is bytes else rettype_func['unicode']

	def vim_get_func(f, rettype=None):
		'''Return a vim function binding.'''
		try:
			func = vim.bindeval('function("' + f + '")')
		except vim.error:
			return None
		else:
			return rettype_func[rettype](func)
else:
	rettype_eval = {
		None: getattr(vim, 'eval', None),
		'int': lambda expr: int(vim.eval(expr)),
		'bytes': eval_bytes,
		'unicode': eval_unicode,
	}
	rettype_eval['str'] = rettype_eval[None]

	class VimFunc(object):
		'''Evaluate a vim function using vim.eval().

		This is a fallback class for older vim versions.
		'''
		__slots__ = ('f', 'eval')

		def __init__(self, f, rettype=None):
			self.f = f.encode('utf-8')
			self.eval = rettype_eval[rettype]

		def __call__(self, *args):
			return self.eval(self.f + b'(' + (b','.join((
				python_to_vim(o) for o in args
			))) + b')')

	vim_get_func = VimFunc


def vim_get_autoload_func(f, rettype=None):
	func = vim_get_func(f)
	if not func:
		vim.command('runtime! ' + f.replace('#', '/')[:f.rindex('#')] + '.vim')
		func = vim_get_func(f)
	return func


if hasattr(vim, 'Function'):
	def vim_func_exists(f):
		try:
			vim.Function(f)
		except ValueError:
			return False
		else:
			return True
else:
	def vim_func_exists(f):
		try:
			return bool(int(vim.eval('exists("*{0}")'.format(f))))
		except vim.error:
			return False


if type(vim) is object:
	vim_get_func = lambda *args, **kwargs: None


_getbufvar = vim_get_func('getbufvar')
_vim_exists = vim_get_func('exists', rettype='int')


# It may crash on some old vim versions and I do not remember in which patch 
# I fixed this crash.
if hasattr(vim, 'vvars') and vim.vvars[str('version')] > 703:
	_vim_to_python_types = {
		getattr(vim, 'Dictionary', None) or type(vim.bindeval('{}')):
			lambda value: dict((
				(_vim_to_python(k), _vim_to_python(v))
				for k, v in value.items()
			)),
		getattr(vim, 'List', None) or type(vim.bindeval('[]')):
			lambda value: [_vim_to_python(item) for item in value],
		getattr(vim, 'Function', None) or type(vim.bindeval('function("mode")')):
			lambda _: None,
	}

	def vim_getvar(varname):
		return _vim_to_python(vim.vars[str(varname)])

	def bufvar_exists(buffer, varname):
		buffer = buffer or vim.current.buffer
		return varname in buffer.vars

	def vim_getwinvar(segment_info, varname):
		return _vim_to_python(segment_info['window'].vars[str(varname)])

	def vim_global_exists(name):
		try:
			vim.vars[name]
		except KeyError:
			return False
		else:
			return True
else:
	_vim_to_python_types = {
		dict: (lambda value: dict(((k, _vim_to_python(v)) for k, v in value.items()))),
		list: (lambda value: [_vim_to_python(i) for i in value]),
	}

	def vim_getvar(varname):
		varname = 'g:' + varname
		if _vim_exists(varname):
			return vim.eval(varname)
		else:
			raise KeyError(varname)

	def bufvar_exists(buffer, varname):
		if not buffer or buffer.number == vim.current.buffer.number:
			return int(vim.eval('exists("b:{0}")'.format(varname)))
		else:
			return int(vim.eval(
				'has_key(getbufvar({0}, ""), {1})'.format(buffer.number, varname)
			))

	def vim_getwinvar(segment_info, varname):
		result = vim.eval('getwinvar({0}, "{1}")'.format(segment_info['winnr'], varname))
		if result == '':
			if not int(vim.eval('has_key(getwinvar({0}, ""), "{1}")'.format(segment_info['winnr'], varname))):
				raise KeyError(varname)
		return result

	def vim_global_exists(name):
		return int(vim.eval('exists("g:' + name + '")'))


def vim_command_exists(name):
	return _vim_exists(':' + name)


if sys.version_info < (3,):
	getbufvar = _getbufvar
else:
	_vim_to_python_types[bytes] = lambda value: value.decode(vim_encoding)

	def getbufvar(*args):
		return _vim_to_python(_getbufvar(*args))


_id = lambda value: value


def _vim_to_python(value):
	return _vim_to_python_types.get(type(value), _id)(value)


if hasattr(vim, 'options'):
	def vim_getbufoption(info, option):
		return info['buffer'].options[str(option)]

	def vim_getoption(option):
		return vim.options[str(option)]

	def vim_setoption(option, value):
		vim.options[str(option)] = value
else:
	def vim_getbufoption(info, option):
		return getbufvar(info['bufnr'], '&' + option)

	def vim_getoption(option):
		return vim.eval('&g:' + option)

	def vim_setoption(option, value):
		vim.command('let &g:{option} = {value}'.format(
			option=option, value=python_to_vim(value)))


if hasattr(vim, 'tabpages'):
	current_tabpage = lambda: vim.current.tabpage
	list_tabpages = lambda: vim.tabpages

	def list_tabpage_buffers_segment_info(segment_info):
		return (
			{'buffer': window.buffer, 'bufnr': window.buffer.number}
			for window in segment_info['tabpage'].windows
		)
else:
	class FalseObject(object):
		@staticmethod
		def __nonzero__():
			return False

		__bool__ = __nonzero__

	def get_buffer(number):
		for buffer in vim.buffers:
			if buffer.number == number:
				return buffer
		raise KeyError(number)

	class WindowVars(object):
		__slots__ = ('tabnr', 'winnr')

		def __init__(self, window):
			self.tabnr = window.tabnr
			self.winnr = window.number

		def __getitem__(self, key):
			has_key = vim.eval('has_key(gettabwinvar({0}, {1}, ""), "{2}")'.format(self.tabnr, self.winnr, key))
			if has_key == '0':
				raise KeyError
			return vim.eval('gettabwinvar({0}, {1}, "{2}")'.format(self.tabnr, self.winnr, key))

		def get(self, key, default=None):
			try:
				return self[key]
			except KeyError:
				return default

	class Window(FalseObject):
		__slots__ = ('tabnr', 'number', '_vars')

		def __init__(self, tabnr, number):
			self.tabnr = tabnr
			self.number = number
			self.vars = WindowVars(self)

		@property
		def buffer(self):
			return get_buffer(int(vim.eval('tabpagebuflist({0})[{1}]'.format(self.tabnr, self.number - 1))))

	class Tabpage(FalseObject):
		__slots__ = ('number',)

		def __init__(self, number):
			self.number = number

		def __eq__(self, tabpage):
			if not isinstance(tabpage, Tabpage):
				raise NotImplementedError
			return self.number == tabpage.number

		@property
		def window(self):
			return Window(self.number, int(vim.eval('tabpagewinnr({0})'.format(self.number))))

	def _last_tab_nr():
		return int(vim.eval('tabpagenr("$")'))

	def current_tabpage():
		return Tabpage(int(vim.eval('tabpagenr()')))

	def list_tabpages():
		return [Tabpage(nr) for nr in range(1, _last_tab_nr() + 1)]

	class TabBufSegmentInfo(dict):
		def __getitem__(self, key):
			try:
				return super(TabBufSegmentInfo, self).__getitem__(key)
			except KeyError:
				if key != 'buffer':
					raise
				else:
					buffer = get_buffer(super(TabBufSegmentInfo, self).__getitem__('bufnr'))
					self['buffer'] = buffer
					return buffer

	def list_tabpage_buffers_segment_info(segment_info):
		return (
			TabBufSegmentInfo(bufnr=int(bufnrstr))
			for bufnrstr in vim.eval('tabpagebuflist({0})'.format(segment_info['tabnr']))
		)


class VimEnviron(object):
	@staticmethod
	def __getitem__(key):
		return vim.eval('$' + key)

	@staticmethod
	def get(key, default=None):
		return vim.eval('$' + key) or default

	@staticmethod
	def __setitem__(key, value):
		return vim.command(
			'let ${0}="{1}"'.format(
				key,
				value.replace('"', '\\"')
				     .replace('\\', '\\\\')
				     .replace('\n', '\\n')
				     .replace('\0', '')
			)
		)


if sys.version_info < (3,):
	def buffer_name(segment_info):
		return segment_info['buffer'].name
else:
	vim_bufname = vim_get_func('bufname', rettype='bytes')

	def buffer_name(segment_info):
		try:
			name = segment_info['buffer'].name
		except UnicodeDecodeError:
			return vim_bufname(segment_info['bufnr'])
		else:
			return name.encode(segment_info['encoding']) if name else None


vim_strtrans = vim_get_func('strtrans', rettype='unicode')


def powerline_vim_strtrans_error(e):
	if not isinstance(e, UnicodeDecodeError):
		raise NotImplementedError
	text = vim_strtrans(e.object[e.start:e.end])
	return (text, e.end)


codecs.register_error('powerline_vim_strtrans_error', powerline_vim_strtrans_error)


did_autocmd = False
buffer_caches = []


def register_buffer_cache(cachedict):
	global did_autocmd
	global buffer_caches
	from powerline.vim import get_default_pycmd, pycmd
	if not did_autocmd:
		import __main__
		__main__.powerline_on_bwipe = on_bwipe
		vim.command('augroup Powerline')
		vim.command('	autocmd! BufWipeout * :{pycmd} powerline_on_bwipe()'.format(
			pycmd=(pycmd or get_default_pycmd())))
		vim.command('augroup END')
		did_autocmd = True
	buffer_caches.append(cachedict)
	return cachedict


def on_bwipe():
	global buffer_caches
	bufnr = int(vim.eval('expand("<abuf>")'))
	for cachedict in buffer_caches:
		cachedict.pop(bufnr, None)


environ = VimEnviron()


def create_ruby_dpowerline():
	vim.command((
		'''
		ruby
		if $powerline == nil
			class Powerline
			end
			$powerline = Powerline.new
		end
		'''
	))



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import time

from threading import Lock
from argparse import ArgumentParser

from powerline import Powerline
from powerline.lib.monotonic import monotonic
from powerline.lib.encoding import get_unicode_writer


class BarPowerline(Powerline):
	get_encoding = staticmethod(lambda: 'utf-8')

	def init(self):
		super(BarPowerline, self).init(ext='wm', renderer_module='bar')


def render(event=None, data=None, sub=None):
	global lock
	with lock:
		write(powerline.render())
		write('\n')
		sys.stdout.flush()


if __name__ == '__main__':
	parser = ArgumentParser(description='Powerline BAR bindings.')
	parser.add_argument(
		'--i3', action='store_true',
		help='Subscribe for i3 events.'
	)
	args = parser.parse_args()
	powerline = BarPowerline()

	interval = 0.5
	lock = Lock()

	write = get_unicode_writer(encoding='utf-8')

	if args.i3:
		import i3
		sub = i3.Subscription(render, 'workspace')

	while True:
		start_time = monotonic()
		render()
		time.sleep(max(interval - (monotonic() - start_time), 0.1))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderer import Renderer
from powerline.colorscheme import ATTR_BOLD, ATTR_ITALIC, ATTR_UNDERLINE


def attrs_to_tmux_attrs(attrs):
	if attrs is False:
		return ['nobold', 'noitalics', 'nounderscore']
	else:
		ret = []
		if attrs & ATTR_BOLD:
			ret += ['bold']
		else:
			ret += ['nobold']
		if attrs & ATTR_ITALIC:
			ret += ['italics']
		else:
			ret += ['noitalics']
		if attrs & ATTR_UNDERLINE:
			ret += ['underscore']
		else:
			ret += ['nounderscore']
		return ret


class TmuxRenderer(Renderer):
	'''Powerline tmux segment renderer.'''

	character_translations = Renderer.character_translations.copy()
	character_translations[ord('#')] = '##[]'

	def hlstyle(self, fg=None, bg=None, attrs=None):
		'''Highlight a segment.'''
		# We don’t need to explicitly reset attributes, so skip those calls
		if not attrs and not bg and not fg:
			return ''
		tmux_attrs = []
		if fg is not None:
			if fg is False or fg[0] is False:
				tmux_attrs += ['fg=default']
			else:
				tmux_attrs += ['fg=colour' + str(fg[0])]
		if bg is not None:
			if bg is False or bg[0] is False:
				tmux_attrs += ['bg=default']
			else:
				tmux_attrs += ['bg=colour' + str(bg[0])]
		if attrs is not None:
			tmux_attrs += attrs_to_tmux_attrs(attrs)
		return '#[' + ','.join(tmux_attrs) + ']'

	def get_segment_info(self, segment_info, mode):
		r = self.segment_info.copy()
		if segment_info:
			r.update(segment_info)
		if 'pane_id' in r:
			varname = 'TMUX_PWD_' + r['pane_id'].lstrip('%')
			if varname in r['environ']:
				r['getcwd'] = lambda: r['environ'][varname]
		r['mode'] = mode
		return r


renderer = TmuxRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import json

from powerline.lib.dict import REMOVE_THIS_KEY


def parse_value(s):
	'''Convert string to Python object

	Rules:

	* Empty string means that corresponding key should be removed from the 
	  dictionary.
	* Strings that start with a minus, digit or with some character that starts 
	  JSON collection or string object are parsed as JSON.
	* JSON special values ``null``, ``true``, ``false`` (case matters) are 
	  parsed  as JSON.
	* All other values are considered to be raw strings.

	:param str s: Parsed string.

	:return: Python object.
	'''
	if not s:
		return REMOVE_THIS_KEY
	elif s[0] in '"{[0193456789-' or s in ('null', 'true', 'false'):
		return json.loads(s)
	else:
		return s


def keyvaluesplit(s):
	'''Split K1.K2=VAL into K1.K2 and parsed VAL
	'''
	if '=' not in s:
		raise TypeError('Option must look like option=json_value')
	if s[0] == '_':
		raise ValueError('Option names must not start with `_\'')
	idx = s.index('=')
	o = s[:idx]
	val = parse_value(s[idx + 1:])
	return (o, val)


def parsedotval(s):
	'''Parse K1.K2=VAL into {"K1":{"K2":VAL}}

	``VAL`` is processed according to rules defined in :py:func:`parse_value`.
	'''
	if type(s) is tuple:
		o, val = s
		val = parse_value(val)
	else:
		o, val = keyvaluesplit(s)

	keys = o.split('.')
	if len(keys) > 1:
		r = (keys[0], {})
		rcur = r[1]
		for key in keys[1:-1]:
			rcur[key] = {}
			rcur = rcur[key]
		rcur[keys[-1]] = val
		return r
	else:
		return (o, val)


def parse_override_var(s):
	'''Parse a semicolon-separated list of strings into a sequence of values

	Emits the same items in sequence as :py:func:`parsedotval` does.
	'''
	return (
		parsedotval(item)
		for item in s.split(';')
		if item
	)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from libqtile import bar
from libqtile.widget import base

from powerline import Powerline as PowerlineCore


class Powerline(base._TextBox):
	def __init__(self, timeout=2, text=' ', width=bar.CALCULATED, **config):
		base._TextBox.__init__(self, text, width, **config)
		self.timeout_add(timeout, self.update)
		self.powerline = PowerlineCore(ext='wm', renderer_module='pango_markup')

	def update(self):
		if not self.configured:
			return True
		self.text = self.powerline.render(side='right')
		self.bar.draw()
		return True

	def cmd_update(self, text):
		self.update(text)

	def cmd_get(self):
		return self.text

	def _configure(self, qtile, bar):
		base._TextBox._configure(self, qtile, bar)
		self.layout = self.drawer.textlayout(
			self.text,
			self.foreground,
			self.font,
			self.fontsize,
			self.fontshadow,
			markup=True)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)


# Abstract classes.
class Event(object):
	def __init__(self, start_mark=None, end_mark=None):
		self.start_mark = start_mark
		self.end_mark = end_mark

	def __repr__(self):
		attributes = [
			key for key in ['implicit', 'value']
			if hasattr(self, key)
		]
		arguments = ', '.join([
			'%s=%r' % (key, getattr(self, key))
			for key in attributes
		])
		return '%s(%s)' % (self.__class__.__name__, arguments)


class NodeEvent(Event):
	def __init__(self, start_mark=None, end_mark=None):
		self.start_mark = start_mark
		self.end_mark = end_mark


class CollectionStartEvent(NodeEvent):
	def __init__(self, implicit, start_mark=None, end_mark=None, flow_style=None):
		self.tag = None
		self.implicit = implicit
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.flow_style = flow_style


class CollectionEndEvent(Event):
	pass


# Implementations.
class StreamStartEvent(Event):
	def __init__(self, start_mark=None, end_mark=None, encoding=None):
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.encoding = encoding


class StreamEndEvent(Event):
	pass


class DocumentStartEvent(Event):
	def __init__(self, start_mark=None, end_mark=None, explicit=None, version=None, tags=None):
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.explicit = explicit
		self.version = version
		self.tags = tags


class DocumentEndEvent(Event):
	def __init__(self, start_mark=None, end_mark=None, explicit=None):
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.explicit = explicit


class AliasEvent(NodeEvent):
	pass


class ScalarEvent(NodeEvent):
	def __init__(self, implicit, value, start_mark=None, end_mark=None, style=None):
		self.tag = None
		self.implicit = implicit
		self.value = value
		self.start_mark = start_mark
		self.end_mark = end_mark
		self.style = style


class SequenceStartEvent(CollectionStartEvent):
	pass


class SequenceEndEvent(CollectionEndEvent):
	pass


class MappingStartEvent(CollectionStartEvent):
	pass


class MappingEndEvent(CollectionEndEvent):
	pass



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.lint.markedjson.loader import Loader


def load(stream, Loader=Loader):
	'''Parse JSON value and produce the corresponding Python object

	:return:
		(hadproblem, object) where first argument is true if there were errors 
		during loading JSON stream and second is the corresponding JSON object.
	'''
	loader = Loader(stream)
	try:
		r = loader.get_single_data()
		return r, loader.haserrors
	finally:
		loader.dispose()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os


POWERLINE_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BINDINGS_DIRECTORY = os.path.join(POWERLINE_ROOT, 'powerline', 'bindings')
TMUX_CONFIG_DIRECTORY = os.path.join(BINDINGS_DIRECTORY, 'tmux')
DEFAULT_SYSTEM_CONFIG_DIR = None



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import re


NON_PRINTABLE_STR = (
	'[^'
	# ASCII control characters: 0x00-0x19
	+ '\t\n'           # Tab, newline: allowed ASCII control characters
	+ '\x20-\x7E'      # ASCII printable characters
	# Unicode control characters: 0x7F-0x9F
	+ '\u0085'         # Allowed unicode control character: next line character
	+ '\u00A0-\uD7FF'
	# Surrogate escapes: 0xD800-0xDFFF
	+ '\uE000-\uFFFD'
	+ ((
		'\uD800-\uDFFF'
	) if sys.maxunicode < 0x10FFFF else (
		'\U00010000-\U0010FFFF'
	))
	+ ']'
	+ ((
		# Paired surrogate escapes: allowed in UCS-2 builds as the only way to 
		# represent characters above 0xFFFF. Only paired variant is allowed.
		'|(?<![\uD800-\uDBFF])[\uDC00-\uDFFF]'
		+ '|[\uD800-\uDBFF](?![\uDC00-\uDFFF])'
	) if sys.maxunicode < 0x10FFFF else (
		''
	))
)
NON_PRINTABLE_RE = re.compile(NON_PRINTABLE_STR)


def repl(s):
	return '<x%04x>' % ord(s.group())


def strtrans(s):
	return NON_PRINTABLE_RE.sub(repl, s.replace('\t', '>---'))


class Mark:
	def __init__(self, name, line, column, buffer, pointer, old_mark=None, merged_marks=None):
		self.name = name
		self.line = line
		self.column = column
		self.buffer = buffer
		self.pointer = pointer
		self.old_mark = old_mark
		self.merged_marks = merged_marks or []

	def copy(self):
		return Mark(self.name, self.line, self.column, self.buffer, self.pointer, self.old_mark, self.merged_marks[:])

	def get_snippet(self, indent=4, max_length=75):
		if self.buffer is None:
			return None
		head = ''
		start = self.pointer
		while start > 0 and self.buffer[start - 1] not in '\0\n':
			start -= 1
			if self.pointer - start > max_length / 2 - 1:
				head = ' ... '
				start += 5
				break
		tail = ''
		end = self.pointer
		while end < len(self.buffer) and self.buffer[end] not in '\0\n':
			end += 1
			if end - self.pointer > max_length / 2 - 1:
				tail = ' ... '
				end -= 5
				break
		snippet = [self.buffer[start:self.pointer], self.buffer[self.pointer], self.buffer[self.pointer + 1:end]]
		snippet = [strtrans(s) for s in snippet]
		return (
			' ' * indent + head + ''.join(snippet) + tail + '\n'
			+ ' ' * (indent + len(head) + len(snippet[0])) + '^'
		)

	def advance_string(self, diff):
		ret = self.copy()
		# FIXME Currently does not work properly with escaped strings.
		ret.column += diff
		ret.pointer += diff
		return ret

	def set_old_mark(self, old_mark):
		if self is old_mark:
			return
		checked_marks = set([id(self)])
		older_mark = old_mark
		while True:
			if id(older_mark) in checked_marks:
				raise ValueError('Trying to set recursive marks')
			checked_marks.add(id(older_mark))
			older_mark = older_mark.old_mark
			if not older_mark:
				break
		self.old_mark = old_mark

	def set_merged_mark(self, merged_mark):
		self.merged_marks.append(merged_mark)

	def to_string(self, indent=0, head_text='in ', add_snippet=True):
		mark = self
		where = ''
		processed_marks = set()
		while mark:
			indentstr = ' ' * indent
			where += ('%s  %s"%s", line %d, column %d' % (
				indentstr, head_text, mark.name, mark.line + 1, mark.column + 1))
			if add_snippet:
				snippet = mark.get_snippet(indent=(indent + 4))
				if snippet:
					where += ':\n' + snippet
			if mark.merged_marks:
				where += '\n' + indentstr + '  with additionally merged\n'
				where += mark.merged_marks[0].to_string(indent + 4, head_text='', add_snippet=False)
				for mmark in mark.merged_marks[1:]:
					where += '\n' + indentstr + '  and\n'
					where += mmark.to_string(indent + 4, head_text='', add_snippet=False)
			if add_snippet:
				processed_marks.add(id(mark))
				if mark.old_mark:
					where += '\n' + indentstr + '  which replaced value\n'
					indent += 4
			mark = mark.old_mark
			if id(mark) in processed_marks:
				raise ValueError('Trying to dump recursive mark')
		if type(where) is str:
			return where
		else:
			return where.encode('utf-8')

	def __str__(self):
		return self.to_string()


def echoerr(*args, **kwargs):
	stream = kwargs.pop('stream', sys.stderr)
	stream.write('\n')
	stream.write(format_error(*args, **kwargs) + '\n')


def format_error(context=None, context_mark=None, problem=None, problem_mark=None, note=None):
	lines = []
	if context is not None:
		lines.append(context)
	if (
		context_mark is not None
		and (
			problem is None or problem_mark is None
			or context_mark.name != problem_mark.name
			or context_mark.line != problem_mark.line
			or context_mark.column != problem_mark.column
		)
	):
		lines.append(str(context_mark))
	if problem is not None:
		lines.append(problem)
	if problem_mark is not None:
		lines.append(str(problem_mark))
	if note is not None:
		lines.append(note)
	return '\n'.join(lines)


class MarkedError(Exception):
	def __init__(self, context=None, context_mark=None, problem=None, problem_mark=None, note=None):
		Exception.__init__(self, format_error(context, context_mark, problem, problem_mark, note))


class EchoErr(object):
	__slots__ = ('echoerr', 'logger',)

	def __init__(self, echoerr, logger):
		self.echoerr = echoerr
		self.logger = logger

	def __call__(self, *args, **kwargs):
		self.echoerr(*args, **kwargs)


class DelayedEchoErr(EchoErr):
	__slots__ = ('echoerr', 'logger', 'errs')

	def __init__(self, echoerr):
		super(DelayedEchoErr, self).__init__(echoerr, echoerr.logger)
		self.errs = []

	def __call__(self, *args, **kwargs):
		self.errs.append((args, kwargs))

	def echo_all(self):
		for args, kwargs in self.errs:
			self.echoerr(*args, **kwargs)

	def __nonzero__(self):
		return not not self.errs

	__bool__ = __nonzero__



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.lint.markedjson import nodes
from powerline.lint.markedjson import events
from powerline.lint.markedjson.error import MarkedError


__all__ = ['Composer', 'ComposerError']


class ComposerError(MarkedError):
	pass


class Composer:
	def __init__(self):
		pass

	def check_node(self):
		# Drop the STREAM-START event.
		if self.check_event(events.StreamStartEvent):
			self.get_event()

		# If there are more documents available?
		return not self.check_event(events.StreamEndEvent)

	def get_node(self):
		# Get the root node of the next document.
		if not self.check_event(events.StreamEndEvent):
			return self.compose_document()

	def get_single_node(self):
		# Drop the STREAM-START event.
		self.get_event()

		# Compose a document if the stream is not empty.
		document = None
		if not self.check_event(events.StreamEndEvent):
			document = self.compose_document()

		# Ensure that the stream contains no more documents.
		if not self.check_event(events.StreamEndEvent):
			event = self.get_event()
			raise ComposerError(
				'expected a single document in the stream',
				document.start_mark,
				'but found another document',
				event.start_mark
			)

		# Drop the STREAM-END event.
		self.get_event()

		return document

	def compose_document(self):
		# Drop the DOCUMENT-START event.
		self.get_event()

		# Compose the root node.
		node = self.compose_node(None, None)

		# Drop the DOCUMENT-END event.
		self.get_event()

		return node

	def compose_node(self, parent, index):
		self.descend_resolver(parent, index)
		if self.check_event(events.ScalarEvent):
			node = self.compose_scalar_node()
		elif self.check_event(events.SequenceStartEvent):
			node = self.compose_sequence_node()
		elif self.check_event(events.MappingStartEvent):
			node = self.compose_mapping_node()
		self.ascend_resolver()
		return node

	def compose_scalar_node(self):
		event = self.get_event()
		tag = event.tag
		if tag is None or tag == '!':
			tag = self.resolve(nodes.ScalarNode, event.value, event.implicit, event.start_mark)
		node = nodes.ScalarNode(tag, event.value, event.start_mark, event.end_mark, style=event.style)
		return node

	def compose_sequence_node(self):
		start_event = self.get_event()
		tag = start_event.tag
		if tag is None or tag == '!':
			tag = self.resolve(nodes.SequenceNode, None, start_event.implicit)
		node = nodes.SequenceNode(tag, [], start_event.start_mark, None, flow_style=start_event.flow_style)
		index = 0
		while not self.check_event(events.SequenceEndEvent):
			node.value.append(self.compose_node(node, index))
			index += 1
		end_event = self.get_event()
		node.end_mark = end_event.end_mark
		return node

	def compose_mapping_node(self):
		start_event = self.get_event()
		tag = start_event.tag
		if tag is None or tag == '!':
			tag = self.resolve(nodes.MappingNode, None, start_event.implicit)
		node = nodes.MappingNode(tag, [], start_event.start_mark, None, flow_style=start_event.flow_style)
		while not self.check_event(events.MappingEndEvent):
			# key_event = self.peek_event()
			item_key = self.compose_node(node, None)
			# if item_key in node.value:
			# 	 raise ComposerError('while composing a mapping', start_event.start_mark,
			# 			 'found duplicate key', key_event.start_mark)
			item_value = self.compose_node(node, item_key)
			# node.value[item_key] = item_value
			node.value.append((item_key, item_value))
		end_event = self.get_event()
		node.end_mark = end_event.end_mark
		return node



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from powerline.theme import requires_segment_info


@requires_segment_info
def current_line(pl, segment_info):
	'''Displays line number that is next to be run
	'''
	return str(segment_info['curframe'].f_lineno)


@requires_segment_info
def current_file(pl, segment_info, basename=True):
	'''Displays current file name

	:param bool basename:
		If true only basename is displayed.
	'''
	filename = segment_info['curframe'].f_code.co_filename
	if basename:
		filename = os.path.basename(filename)
	return filename


@requires_segment_info
def current_code_name(pl, segment_info):
	'''Displays name of the code object of the current frame
	'''
	return segment_info['curframe'].f_code.co_name


@requires_segment_info
def current_context(pl, segment_info):
	'''Displays currently executed context name

	This is similar to :py:func:`current_code_name`, but gives more details.

	Currently it only gives module file name if code_name happens to be 
	``<module>``.
	'''
	name = segment_info['curframe'].f_code.co_name
	if name == '<module>':
		name = os.path.basename(segment_info['curframe'].f_code.co_filename)
	return name


@requires_segment_info
def stack_depth(pl, segment_info, full_stack=False):
	'''Displays current stack depth

	Result is relative to the stack depth at the time prompt was first run.

	:param bool full_stack:
		If true then absolute depth is used.
	'''
	return str(len(segment_info['pdb'].stack) - (
		0 if full_stack else segment_info['initial_stack_length']))



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import socket
import errno
import os

try:
	from posix import environ
except ImportError:
	from os import environ

# XXX Hack for importing powerline modules to work.
sys.path.pop(0)

try:
	from powerline.lib.encoding import get_preferred_output_encoding
except ImportError:
	sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(os.path.realpath(__file__)))))
	from powerline.lib.encoding import get_preferred_output_encoding


if len(sys.argv) < 2:
	print('Must provide at least one argument.', file=sys.stderr)
	raise SystemExit(1)

use_filesystem = not sys.platform.lower().startswith('linux')

if sys.argv[1] == '--socket':
	address = sys.argv[2]
	if not use_filesystem:
		address = '\0' + address
	del sys.argv[1:3]
else:
	address = ('/tmp/powerline-ipc-%d' if use_filesystem else '\0powerline-ipc-%d') % os.getuid()

sock = socket.socket(family=socket.AF_UNIX)


def eintr_retry_call(func, *args, **kwargs):
	while True:
		try:
			return func(*args, **kwargs)
		except EnvironmentError as e:
			if getattr(e, 'errno', None) == errno.EINTR:
				continue
			raise


try:
	eintr_retry_call(sock.connect, address)
except Exception:
	# Run the powerline renderer
	args = ['powerline-render'] + sys.argv[1:]
	os.execvp('powerline-render', args)

fenc = get_preferred_output_encoding()


def tobytes(s):
	if isinstance(s, bytes):
		return s
	else:
		return s.encode(fenc)


args = [tobytes('%x' % (len(sys.argv) - 1))]
args.extend((tobytes(s) for s in sys.argv[1:]))


try:
	cwd = os.getcwd()
except EnvironmentError:
	pass
else:
	if not isinstance(cwd, bytes):
		cwd = cwd.encode(fenc)
	args.append(cwd)


args.extend((tobytes(k) + b'=' + tobytes(v) for k, v in environ.items()))

EOF = b'\0\0'

for a in args:
	eintr_retry_call(sock.sendall, a + b'\0')

eintr_retry_call(sock.sendall, EOF)

received = []
while True:
	r = sock.recv(4096)
	if not r:
		break
	received.append(r)

sock.close()

if sys.version_info < (3,):
	sys.stdout.write(b''.join(received))
else:
	sys.stdout.buffer.write(b''.join(received))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

try:
	from urllib.error import HTTPError  # NOQA
	from urllib.request import urlopen  # NOQA
	from urllib.parse import urlencode as urllib_urlencode  # NOQA
except ImportError:
	from urllib2 import urlopen, HTTPError  # NOQA
	from urllib import urlencode as urllib_urlencode  # NOQA


def urllib_read(url):
	try:
		return urlopen(url, timeout=10).read().decode('utf-8')
	except HTTPError:
		return



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys

from powerline.lib.shell import asrun, run_cmd
from powerline.lib.unicode import out_u
from powerline.segments import Segment, with_docstring


STATE_SYMBOLS = {
	'fallback': '',
	'play': '>',
	'pause': '~',
	'stop': 'X',
}


def _convert_state(state):
	'''Guess player state'''
	state = state.lower()
	if 'play' in state:
		return 'play'
	if 'pause' in state:
		return 'pause'
	if 'stop' in state:
		return 'stop'
	return 'fallback'


def _convert_seconds(seconds):
	'''Convert seconds to minutes:seconds format'''
	return '{0:.0f}:{1:02.0f}'.format(*divmod(float(seconds), 60))


class PlayerSegment(Segment):
	def __call__(self, format='{state_symbol} {artist} - {title} ({total})', state_symbols=STATE_SYMBOLS, **kwargs):
		stats = {
			'state': 'fallback',
			'album': None,
			'artist': None,
			'title': None,
			'elapsed': None,
			'total': None,
		}
		func_stats = self.get_player_status(**kwargs)
		if not func_stats:
			return None
		stats.update(func_stats)
		stats['state_symbol'] = state_symbols.get(stats['state'])
		return [{
			'contents': format.format(**stats),
			'highlight_groups': ['player_' + (stats['state'] or 'fallback'), 'player'],
		}]

	def get_player_status(self, pl):
		pass

	def argspecobjs(self):
		for ret in super(PlayerSegment, self).argspecobjs():
			yield ret
		yield 'get_player_status', self.get_player_status

	def omitted_args(self, name, method):
		return (0,)


_common_args = '''
This player segment should be added like this:

.. code-block:: json

	{{
		"function": "powerline.segments.common.players.{0}",
		"name": "player"
	}}

(with additional ``"args": {{…}}`` if needed).

Highlight groups used: ``player_fallback`` or ``player``, ``player_play`` or ``player``, ``player_pause`` or ``player``, ``player_stop`` or ``player``.

:param str format:
	Format used for displaying data from player. Should be a str.format-like 
	string with the following keyword parameters:

	+------------+-------------------------------------------------------------+
	|Parameter   |Description                                                  |
	+============+=============================================================+
	|state_symbol|Symbol displayed for play/pause/stop states. There is also   |
	|            |“fallback” state used in case function failed to get player  |
	|            |state. For this state symbol is by default empty. All        |
	|            |symbols are defined in ``state_symbols`` argument.           |
	+------------+-------------------------------------------------------------+
	|album       |Album that is currently played.                              |
	+------------+-------------------------------------------------------------+
	|artist      |Artist whose song is currently played                        |
	+------------+-------------------------------------------------------------+
	|title       |Currently played composition.                                |
	+------------+-------------------------------------------------------------+
	|elapsed     |Composition duration in format M:SS (minutes:seconds).       |
	+------------+-------------------------------------------------------------+
	|total       |Composition length in format M:SS.                           |
	+------------+-------------------------------------------------------------+
:param dict state_symbols:
	Symbols used for displaying state. Must contain all of the following keys:

	========  ========================================================
	Key       Description
	========  ========================================================
	play      Displayed when player is playing.
	pause     Displayed when player is paused.
	stop      Displayed when player is not playing anything.
	fallback  Displayed if state is not one of the above or not known.
	========  ========================================================
'''


_player = with_docstring(PlayerSegment(), _common_args.format('_player'))


class CmusPlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		'''Return cmus player information.

		cmus-remote -Q returns data with multi-level information i.e.
			status playing
			file <file_name>
			tag artist <artist_name>
			tag title <track_title>
			tag ..
			tag n
			set continue <true|false>
			set repeat <true|false>
			set ..
			set n

		For the information we are looking for we don’t really care if we’re on
		the tag level or the set level. The dictionary comprehension in this
		method takes anything in ignore_levels and brings the key inside that
		to the first level of the dictionary.
		'''
		now_playing_str = run_cmd(pl, ['cmus-remote', '-Q'])
		if not now_playing_str:
			return
		ignore_levels = ('tag', 'set',)
		now_playing = dict(((token[0] if token[0] not in ignore_levels else token[1],
			(' '.join(token[1:]) if token[0] not in ignore_levels else
			' '.join(token[2:]))) for token in [line.split(' ') for line in now_playing_str.split('\n')[:-1]]))
		state = _convert_state(now_playing.get('status'))
		return {
			'state': state,
			'album': now_playing.get('album'),
			'artist': now_playing.get('artist'),
			'title': now_playing.get('title'),
			'elapsed': _convert_seconds(now_playing.get('position', 0)),
			'total': _convert_seconds(now_playing.get('duration', 0)),
		}


cmus = with_docstring(CmusPlayerSegment(),
('''Return CMUS player information

Requires cmus-remote command be acessible from $PATH.

{0}
''').format(_common_args.format('cmus')))


class MpdPlayerSegment(PlayerSegment):
	def get_player_status(self, pl, host='localhost', port=6600):
		try:
			import mpd
		except ImportError:
			now_playing = run_cmd(pl, [
				'mpc', 'current',
				'-f', '%album%\n%artist%\n%title%\n%time%',
				'-h', str(host),
				'-p', str(port)
			], strip=False)
			if not now_playing:
				return
			now_playing = now_playing.split('\n')
			return {
				'album': now_playing[0],
				'artist': now_playing[1],
				'title': now_playing[2],
				'total': now_playing[3],
			}
		else:
			client = mpd.MPDClient()
			client.connect(host, port)
			now_playing = client.currentsong()
			if not now_playing:
				return
			status = client.status()
			client.close()
			client.disconnect()
			return {
				'state': status.get('state'),
				'album': now_playing.get('album'),
				'artist': now_playing.get('artist'),
				'title': now_playing.get('title'),
				'elapsed': _convert_seconds(now_playing.get('elapsed', 0)),
				'total': _convert_seconds(now_playing.get('time', 0)),
			}


mpd = with_docstring(MpdPlayerSegment(),
('''Return Music Player Daemon information

Requires mpc command to be acessible from $PATH or ``mpd`` Python module.

{0}
:param str host:
	Host on which mpd runs.
:param int port:
	Port which should be connected to.
''').format(_common_args.format('mpd')))


try:
	import dbus
except ImportError:
	def _get_dbus_player_status(pl, player_name, **kwargs):
		pl.error('Could not add {0} segment: requires dbus module', player_name)
		return
else:
	def _get_dbus_player_status(pl, bus_name, player_path, iface_prop,
	                            iface_player, player_name='player'):
		bus = dbus.SessionBus()
		try:
			player = bus.get_object(bus_name, player_path)
			iface = dbus.Interface(player, iface_prop)
			info = iface.Get(iface_player, 'Metadata')
			status = iface.Get(iface_player, 'PlaybackStatus')
		except dbus.exceptions.DBusException:
			return
		if not info:
			return
		album = info.get('xesam:album')
		title = info.get('xesam:title')
		artist = info.get('xesam:artist')
		state = _convert_state(status)
		if album:
			album = out_u(album)
		if title:
			title = out_u(title)
		if artist:
			artist = out_u(artist[0])
		return {
			'state': state,
			'album': album,
			'artist': artist,
			'title': title,
			'total': _convert_seconds(info.get('mpris:length') / 1e6),
		}


class DbusPlayerSegment(PlayerSegment):
	get_player_status = staticmethod(_get_dbus_player_status)


dbus_player = with_docstring(DbusPlayerSegment(),
('''Return generic dbus player state

Requires ``dbus`` python module. Only for players that support specific protocol 
 (e.g. like :py:func:`spotify` and :py:func:`clementine`).

{0}
:param str player_name:
	Player name. Used in error messages only.
:param str bus_name:
	Dbus bus name.
:param str player_path:
	Path to the player on the given bus.
:param str iface_prop:
	Interface properties name for use with dbus.Interface.
:param str iface_player:
	Player name.
''').format(_common_args.format('dbus_player')))


class SpotifyDbusPlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		return _get_dbus_player_status(
			pl=pl,
			player_name='Spotify',
			bus_name='com.spotify.qt',
			player_path='/',
			iface_prop='org.freedesktop.DBus.Properties',
			iface_player='org.freedesktop.MediaPlayer2',
		)


spotify_dbus = with_docstring(SpotifyDbusPlayerSegment(),
('''Return spotify player information

Requires ``dbus`` python module.

{0}
''').format(_common_args.format('spotify_dbus')))


class SpotifyAppleScriptPlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		status_delimiter = '-~`/='
		ascript = '''
			tell application "System Events"
				set process_list to (name of every process)
			end tell

			if process_list contains "Spotify" then
				tell application "Spotify"
					if player state is playing or player state is paused then
						set track_name to name of current track
						set artist_name to artist of current track
						set album_name to album of current track
						set track_length to duration of current track
						set now_playing to "" & player state & "{0}" & album_name & "{0}" & artist_name & "{0}" & track_name & "{0}" & track_length
						return now_playing
					else
						return player state
					end if

				end tell
			else
				return "stopped"
			end if
		'''.format(status_delimiter)

		spotify = asrun(pl, ascript)
		if not asrun:
			return None

		spotify_status = spotify.split(status_delimiter)
		state = _convert_state(spotify_status[0])
		if state == 'stop':
			return None
		return {
			'state': state,
			'album': spotify_status[1],
			'artist': spotify_status[2],
			'title': spotify_status[3],
			'total': _convert_seconds(int(spotify_status[4]))
		}


spotify_apple_script = with_docstring(SpotifyAppleScriptPlayerSegment(),
('''Return spotify player information

Requires ``osascript`` available in $PATH.

{0}
''').format(_common_args.format('spotify_apple_script')))


if 'dbus' in globals() or not sys.platform.startswith('darwin'):
	spotify = spotify_dbus
	_old_name = 'spotify_dbus'
else:
	spotify = spotify_apple_script
	_old_name = 'spotify_apple_script'


spotify = with_docstring(spotify, spotify.__doc__.replace(_old_name, 'spotify'))


class ClementinePlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		return _get_dbus_player_status(
			pl=pl,
			player_name='Clementine',
			bus_name='org.mpris.MediaPlayer2.clementine',
			player_path='/org/mpris/MediaPlayer2',
			iface_prop='org.freedesktop.DBus.Properties',
			iface_player='org.mpris.MediaPlayer2.Player',
		)


clementine = with_docstring(ClementinePlayerSegment(),
('''Return clementine player information

Requires ``dbus`` python module.

{0}
''').format(_common_args.format('clementine')))


class RhythmboxPlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		now_playing = run_cmd(pl, [
			'rhythmbox-client',
			'--no-start', '--no-present',
			'--print-playing-format', '%at\n%aa\n%tt\n%te\n%td'
		], strip=False)
		if not now_playing:
			return
		now_playing = now_playing.split('\n')
		return {
			'album': now_playing[0],
			'artist': now_playing[1],
			'title': now_playing[2],
			'elapsed': now_playing[3],
			'total': now_playing[4],
		}


rhythmbox = with_docstring(RhythmboxPlayerSegment(),
('''Return rhythmbox player information

Requires ``rhythmbox-client`` available in $PATH.

{0}
''').format(_common_args.format('rhythmbox')))


class RDIOPlayerSegment(PlayerSegment):
	def get_player_status(self, pl):
		status_delimiter = '-~`/='
		ascript = '''
			tell application "System Events"
				set rdio_active to the count(every process whose name is "Rdio")
				if rdio_active is 0 then
					return
				end if
			end tell
			tell application "Rdio"
				set rdio_name to the name of the current track
				set rdio_artist to the artist of the current track
				set rdio_album to the album of the current track
				set rdio_duration to the duration of the current track
				set rdio_state to the player state
				set rdio_elapsed to the player position
				return rdio_name & "{0}" & rdio_artist & "{0}" & rdio_album & "{0}" & rdio_elapsed & "{0}" & rdio_duration & "{0}" & rdio_state
			end tell
		'''.format(status_delimiter)
		now_playing = asrun(pl, ascript)
		if not now_playing:
			return
		now_playing = now_playing.split(status_delimiter)
		if len(now_playing) != 6:
			return
		state = _convert_state(now_playing[5])
		total = _convert_seconds(now_playing[4])
		elapsed = _convert_seconds(float(now_playing[3]) * float(now_playing[4]) / 100)
		return {
			'title': now_playing[0],
			'artist': now_playing[1],
			'album': now_playing[2],
			'elapsed': elapsed,
			'total': total,
			'state': state,
		}


rdio = with_docstring(RDIOPlayerSegment(),
('''Return rdio player information

Requires ``osascript`` available in $PATH.

{0}
''').format(_common_args.format('rdio')))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import i3


def calcgrp(w):
	group = []
	if w['focused']:
		group.append('w_focused')
	if w['urgent']:
		group.append('w_urgent')
	if w['visible']:
		group.append('w_visible')
	group.append('workspace')
	return group


def workspaces(pl):
	'''Return workspace list

	Highlight groups used: ``workspace``, ``w_visible``, ``w_focused``, ``w_urgent``
	'''
	return [{
		'contents': w['name'],
		'highlight_groups': calcgrp(w)
	} for w in i3.get_workspaces()]



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import pdb

from powerline.pdb import PDBPowerline
from powerline.lib.encoding import get_preferred_output_encoding
from powerline.lib.unicode import unicode


if sys.version_info < (3,):
	# XXX The below classes make code compatible with PDBpp which uses pyrepl
	#     which does not expect unicode or something above ASCII. They are 
	#     completely not needed if pdbpp is not used, but that’s not always the 
	#     case.
	class PowerlineRenderBytesResult(bytes):
		def __new__(cls, s, encoding=None):
			encoding = encoding or s.encoding
			if isinstance(s, PowerlineRenderResult):
				return s.encode(encoding)
			self = bytes.__new__(cls, s.encode(encoding) if isinstance(s, unicode) else s)
			self.encoding = encoding
			return self

		for meth in (
			'__contains__',
			'partition', 'rpartition',
			'split', 'rsplit',
			'count', 'join',
		):
			exec((
				'def {0}(self, *args):\n'
				'	if any((isinstance(arg, unicode) for arg in args)):\n'
				'		return self.__unicode__().{0}(*args)\n'
				'	else:\n'
				'		return bytes.{0}(self, *args)'
			).format(meth))

		for meth in (
			'find', 'rfind',
			'index', 'rindex',
		):
			exec((
				'def {0}(self, *args):\n'
				'	if any((isinstance(arg, unicode) for arg in args)):\n'
				'		args = [arg.encode(self.encoding) if isinstance(arg, unicode) else arg for arg in args]\n'
				'	return bytes.{0}(self, *args)'
			).format(meth))

		def __len__(self):
			return len(self.decode(self.encoding))

		def __getitem__(self, *args):
			return PowerlineRenderBytesResult(bytes.__getitem__(self, *args), encoding=self.encoding)

		def __getslice__(self, *args):
			return PowerlineRenderBytesResult(bytes.__getslice__(self, *args), encoding=self.encoding)

		@staticmethod
		def add(encoding, *args):
			if any((isinstance(arg, unicode) for arg in args)):
				return PowerlineRenderResult(''.join((
					arg
					if isinstance(arg, unicode)
					else arg.decode(encoding)
					for arg in args
				)), encoding)
			else:
				return PowerlineRenderBytesResult(b''.join(args), encoding=encoding)

		def __add__(self, other):
			return self.add(self.encoding, self, other)

		def __radd__(self, other):
			return self.add(self.encoding, other, self)

		def __unicode__(self):
			return PowerlineRenderResult(self)

	class PowerlineRenderResult(unicode):
		def __new__(cls, s, encoding=None):
			encoding = (
				encoding
				or getattr(s, 'encoding', None)
				or get_preferred_output_encoding()
			)
			if isinstance(s, unicode):
				self = unicode.__new__(cls, s)
			else:
				self = unicode.__new__(cls, s, encoding, 'replace')
			self.encoding = encoding
			return self

		def __str__(self):
			return PowerlineRenderBytesResult(self)

		def __getitem__(self, *args):
			return PowerlineRenderResult(unicode.__getitem__(self, *args))

		def __getslice__(self, *args):
			return PowerlineRenderResult(unicode.__getslice__(self, *args))

		@staticmethod
		def add(encoding, *args):
			return PowerlineRenderResult(''.join((
				arg
				if isinstance(arg, unicode)
				else arg.decode(encoding)
				for arg in args
			)), encoding)

		def __add__(self, other):
			return self.add(self.encoding, self, other)

		def __radd__(self, other):
			return self.add(self.encoding, other, self)

		def encode(self, *args, **kwargs):
			return PowerlineRenderBytesResult(unicode.encode(self, *args, **kwargs), args[0])
else:
	PowerlineRenderResult = str


def use_powerline_prompt(cls):
	'''Decorator that installs powerline prompt to the class

	:param pdb.Pdb cls:
		Class that should be decorated.

	:return:
		``cls`` argument or a class derived from it. Latter is used to turn 
		old-style classes into new-style classes.
	'''
	@property
	def prompt(self):
		try:
			powerline = self.powerline
		except AttributeError:
			powerline = PDBPowerline()
			powerline.setup(self)
			self.powerline = powerline
		return PowerlineRenderResult(powerline.render(side='left'))

	@prompt.setter
	def prompt(self, _):
		pass

	@prompt.deleter
	def prompt(self):
		pass

	if not hasattr(cls, '__class__'):
		# Old-style class: make it new-style or @property will not work.
		old_cls = cls

		class cls(cls, object):
			__module__ = cls.__module__
			__doc__ = cls.__doc__

		cls.__name__ = old_cls.__name__

	cls.prompt = prompt

	return cls


def main():
	'''Run module as a script

	Uses :py:func:`pdb.main` function directly, but prior to that it mocks 
	:py:class:`pdb.Pdb` class with powerline-specific class instance.
	'''
	orig_pdb = pdb.Pdb

	@use_powerline_prompt
	class Pdb(pdb.Pdb, object):
		def __init__(self):
			orig_pdb.__init__(self)

	pdb.Pdb = Pdb

	return pdb.main()



# vim:fileencoding=utf-8:noet
from __future__ import (division, absolute_import, print_function)

import argparse


def get_argparser(ArgumentParser=argparse.ArgumentParser):
	parser = ArgumentParser(description='Powerline configuration checker.')
	parser.add_argument(
		'-p', '--config-path', action='append', metavar='PATH',
		help='Paths where configuration should be checked, in order. You must '
		     'supply all paths necessary for powerline to work, '
		     'checking partial (e.g. only user overrides) configuration '
		     'is not supported.'
	)
	parser.add_argument(
		'-d', '--debug', action='store_const', const=True,
		help='Display additional information. Used for debugging '
		     '`powerline-lint\' itself, not for debugging configuration.'
	)
	return parser



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import itertools

from powerline.lib.unicode import unicode
from powerline.lint.markedjson.markedvalue import MarkedUnicode
from powerline.lint.selfcheck import havemarks


class JStr(unicode):
	def join(self, iterable):
		return super(JStr, self).join((unicode(item) for item in iterable))


key_sep = JStr('/')


def list_themes(data, context):
	theme_type = data['theme_type']
	ext = data['ext']
	main_theme_name = data['main_config'].get('ext', {}).get(ext, {}).get('theme', None)
	is_main_theme = (data['theme'] == main_theme_name)
	if theme_type == 'top':
		return list(itertools.chain(*[
			[(theme_ext, theme) for theme in theme_configs.values()]
			for theme_ext, theme_configs in data['theme_configs'].items()
		]))
	elif theme_type == 'main' or is_main_theme:
		return [(ext, theme) for theme in data['ext_theme_configs'].values()]
	else:
		return [(ext, context[0][1])]


class Context(tuple):
	for func in dir(tuple):
		if func in ('__getitem__', '__init__', '__getattribute__', '__len__', '__iter__'):
			continue
		exec((
			'def {0}(self, *args, **kwargs):\n'
			'	raise TypeError("{0} is not allowed for Context")'
		).format(func))
	del func

	__slots__ = ()

	def __new__(cls, base, context_key=None, context_value=None):
		if context_key is not None:
			assert(context_value is not None)
			assert(type(base) is Context)
			havemarks(context_key, context_value)
			return tuple.__new__(cls, tuple.__add__(base, ((context_key, context_value),)))
		else:
			havemarks(base)
			return tuple.__new__(cls, ((MarkedUnicode('', base.mark), base),))

	@property
	def key(self):
		return key_sep.join((c[0] for c in self))

	def enter_key(self, value, key):
		return self.enter(value.keydict[key], value[key])

	def enter_item(self, name, item):
		return self.enter(MarkedUnicode(name, item.mark), item)

	def enter(self, context_key, context_value):
		return Context.__new__(Context, self, context_key, context_value)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import codecs

from unicodedata import east_asian_width, combining

from powerline.lib.encoding import get_preferred_output_encoding


try:
	from __builtin__ import unicode
except ImportError:
	unicode = str


try:
	from __builtin__ import unichr
except ImportError:
	unichr = chr


if sys.maxunicode < 0x10FFFF:
	_unichr = unichr

	def unichr(ch):
		if ch <= sys.maxunicode:
			return _unichr(ch)
		else:
			ch -= 0x10000
			return _unichr((ch >> 10) + 0xD800) + _unichr((ch & ((1 << 10) - 1)) + 0xDC00)


def u(s):
	'''Return unicode instance assuming UTF-8 encoded string.
	'''
	if type(s) is unicode:
		return s
	else:
		return unicode(s, 'utf-8')


if sys.version_info < (3,):
	def tointiter(s):
		'''Convert a byte string to the sequence of integers
		'''
		return (ord(c) for c in s)
else:
	def tointiter(s):
		'''Convert a byte string to the sequence of integers
		'''
		return iter(s)


def powerline_decode_error(e):
	if not isinstance(e, UnicodeDecodeError):
		raise NotImplementedError
	return (''.join((
		'<{0:02X}>'.format(c)
		for c in tointiter(e.object[e.start:e.end])
	)), e.end)


codecs.register_error('powerline_decode_error', powerline_decode_error)


last_swe_idx = 0


def register_strwidth_error(strwidth):
	'''Create new encode errors handling method similar to ``replace``

	Like ``replace`` this method uses question marks in place of the characters 
	that cannot be represented in the requested encoding. Unlike ``replace`` the 
	amount of question marks is identical to the amount of display cells 
	offending character occupies. Thus encoding ``…`` (U+2026, HORIZONTAL 
	ELLIPSIS) to ``latin1`` will emit one question mark, but encoding ``Ａ`` 
	(U+FF21, FULLWIDTH LATIN CAPITAL LETTER A) will emit two question marks.

	Since width of some characters depends on the terminal settings and 
	powerline knows how to respect them a single error handling method cannot be 
	used. Instead of it the generator function is used which takes ``strwidth`` 
	function (function that knows how to compute string width respecting all 
	needed settings) and emits new error handling method name.

	:param function strwidth:
		Function that computs string width measured in display cells the string 
		occupies when displayed.

	:return: New error handling method name.
	'''
	global last_swe_idx
	last_swe_idx += 1

	def powerline_encode_strwidth_error(e):
		if not isinstance(e, UnicodeEncodeError):
			raise NotImplementedError
		return ('?' * strwidth(e.object[e.start:e.end]), e.end)

	ename = 'powerline_encode_strwidth_error_{0}'.format(last_swe_idx)
	codecs.register_error(ename, powerline_encode_strwidth_error)
	return ename


def out_u(s):
	'''Return unicode string suitable for displaying

	Unlike other functions assumes get_preferred_output_encoding() first. Unlike 
	u() does not throw exceptions for invalid unicode strings. Unlike 
	safe_unicode() does throw an exception if object is not a string.
	'''
	if isinstance(s, unicode):
		return s
	elif isinstance(s, bytes):
		return unicode(s, get_preferred_output_encoding(), 'powerline_decode_error')
	else:
		raise TypeError('Expected unicode or bytes instance, got {0}'.format(repr(type(s))))


def safe_unicode(s):
	'''Return unicode instance without raising an exception.

	Order of assumptions:
	* ASCII string or unicode object
	* UTF-8 string
	* Object with __str__() or __repr__() method that returns UTF-8 string or 
	  unicode object (depending on python version)
	* String in powerline.lib.encoding.get_preferred_output_encoding() encoding
	* If everything failed use safe_unicode on last exception with which 
	  everything failed
	'''
	try:
		try:
			if type(s) is bytes:
				return unicode(s, 'ascii')
			else:
				return unicode(s)
		except UnicodeDecodeError:
			try:
				return unicode(s, 'utf-8')
			except TypeError:
				return unicode(str(s), 'utf-8')
			except UnicodeDecodeError:
				return unicode(s, get_preferred_output_encoding())
	except Exception as e:
		return safe_unicode(e)


class FailedUnicode(unicode):
	'''Builtin ``unicode`` subclass indicating fatal error

	If your code for some reason wants to determine whether `.render()` method 
	failed it should check returned string for being a FailedUnicode instance. 
	Alternatively you could subclass Powerline and override `.render()` method 
	to do what you like in place of catching the exception and returning 
	FailedUnicode.
	'''
	pass


if sys.version_info < (3,):
	def string(s):
		if type(s) is not str:
			return s.encode('utf-8')
		else:
			return s
else:
	def string(s):
		if type(s) is not str:
			return s.decode('utf-8')
		else:
			return s


string.__doc__ = (
	'''Transform ``unicode`` or ``bytes`` object into ``str`` object

	On Python-2 this encodes ``unicode`` to ``bytes`` (which is ``str``) using 
	UTF-8 encoding; on Python-3 this decodes ``bytes`` to ``unicode`` (which is 
	``str``) using UTF-8 encoding.

	Useful for functions that expect an ``str`` object in both unicode versions, 
	not caring about the semantic differences between them in Python-2 and 
	Python-3.
	'''
)


def surrogate_pair_to_character(high, low):
	'''Transform a pair of surrogate codepoints to one codepoint
	'''
	return 0x10000 + ((high - 0xD800) << 10) + (low - 0xDC00)


_strwidth_documentation = (
	'''Compute string width in display cells

	{0}

	:param dict width_data:
		Dictionary which maps east_asian_width property values to strings 
		lengths. It is expected to contain the following keys and values (from 
		`East Asian Width annex <http://www.unicode.org/reports/tr11/>`_):

		===  ======  ===========================================================
		Key  Value   Description
		===  ======  ===========================================================
		F    2       Fullwidth: all characters that are defined as Fullwidth in 
		             the Unicode Standard [Unicode] by having a compatibility 
		             decomposition of type <wide> to characters elsewhere in the 
		             Unicode Standard that are implicitly narrow but unmarked.
		H    1       Halfwidth: all characters that are explicitly defined as 
		             Halfwidth in the Unicode Standard by having a compatibility 
		             decomposition of type <narrow> to characters elsewhere in 
		             the Unicode Standard that are implicitly wide but unmarked, 
		             plus U+20A9 ₩ WON SIGN.
		W    2       Wide: all other characters that are always wide. These 
		             characters occur only in the context of East Asian 
		             typography where they are wide characters (such as the 
		             Unified Han Ideographs or Squared Katakana Symbols). This 
		             category includes characters that have explicit halfwidth 
		             counterparts.
		Na   1       Narrow: characters that are always narrow and have explicit 
		             fullwidth or wide counterparts. These characters are 
		             implicitly narrow in East Asian typography and legacy 
		             character sets because they have explicit fullwidth or wide 
		             counterparts. All of ASCII is an example of East Asian 
		             Narrow characters.
		A    1 or 2  Ambigious: characters that may sometimes be wide and 
		             sometimes narrow. Ambiguous characters require additional 
		             information not contained in the character code to further 
		             resolve their width. This information is usually defined in 
		             terminal setting that should in turn respect glyphs widths 
		             in used fonts. Also see :ref:`ambiwidth configuration 
		             option <config-common-ambiwidth>`.
		N    1       Neutral characters: character that does not occur in legacy 
		             East Asian character sets.
		===  ======  ===========================================================

	:param unicode string:
		String whose width will be calculated.

	:return: unsigned integer.''')


def strwidth_ucs_4(width_data, string):
	return sum(((
		(
			0
		) if combining(symbol) else (
			width_data[east_asian_width(symbol)]
		)
	) for symbol in string))


strwidth_ucs_4.__doc__ = _strwidth_documentation.format(
	'''This version of function expects that characters above 0xFFFF are 
	represented using one symbol. This is only the case in UCS-4 Python builds.

	.. note:
		Even in UCS-4 Python builds it is possible to represent characters above 
		0xFFFF using surrogate pairs. Characters represented this way are not 
		supported.''')


def strwidth_ucs_2(width_data, string):
	return sum(((
		(
			width_data[east_asian_width(string[i - 1] + symbol)]
		) if 0xDC00 <= ord(symbol) <= 0xDFFF else (
			0
		) if combining(symbol) or 0xD800 <= ord(symbol) <= 0xDBFF else (
			width_data[east_asian_width(symbol)]
		)
	) for i, symbol in enumerate(string)))


strwidth_ucs_2.__doc__ = _strwidth_documentation.format(
	'''This version of function expects that characters above 0xFFFF are 
	represented using two symbols forming a surrogate pair, which is the only 
	option in UCS-2 Python builds. It still works correctly in UCS-4 Python 
	builds, but is slower then its UCS-4 counterpart.''')
